# Licensed under a 3-clause BSD style license - see LICENSE.rst

# This module includes files automatically generated from ply (these end in
# _lextab.py and _parsetab.py). To generate these files, remove them from this
# folder, then build astropy and run the tests in-place:
#
#   python setup.py build_ext --inplace
#   pytest astropy/units
#
# You can then commit the changes to the re-generated _lextab.py and
# _parsetab.py files.

"""
Handles a "generic" string format for units
"""

import re
import unicodedata
import warnings
from fractions import Fraction
from re import Match, Pattern
from typing import ClassVar, Final

import numpy as np

from astropy.extern.ply.lex import Lexer
from astropy.units.core import CompositeUnit, Unit, UnitBase, get_current_unit_registry
from astropy.units.enums import DeprecatedUnitAction
from astropy.units.errors import UnitsWarning
from astropy.units.typing import UnitScale
from astropy.utils import classproperty, parsing
from astropy.utils.parsing import ThreadSafeParser

from .base import Base, _ParsingFormatMixin


class _GenericParserMixin(_ParsingFormatMixin):
    """Provide the parser used by Generic, FITS and VOUnit."""

    _tokens: ClassVar[tuple[str, ...]] = (
        "COMMA",
        "POWER",
        "PRODUCT",
        "DIVISION",
        "OPEN_PAREN",
        "CLOSE_PAREN",
        "FUNCNAME",
        "UNIT",
        "SIGN",
        "UINT",
        "UFLOAT",
    )

    @classproperty(lazy=True)
    def _lexer(cls) -> Lexer:
        tokens = cls._tokens

        t_COMMA = r"\,"
        t_PRODUCT = "[*.]"
        t_DIVISION = "/"
        t_POWER = r"\^|(\*\*)"
        t_OPEN_PAREN = r"\("
        t_CLOSE_PAREN = r"\)"

        # NOTE THE ORDERING OF THESE RULES IS IMPORTANT!!
        # Regular expression rules for simple tokens
        def t_UFLOAT(t):
            r"((\d+\.?\d*)|(\.\d+))([eE][+-]?\d+)?"
            if not re.search(r"[eE\.]", t.value):
                t.type = "UINT"
                t.value = int(t.value)
            elif t.value.endswith("."):
                t.type = "UINT"
                t.value = int(t.value[:-1])
            else:
                t.value = float(t.value)
            return t

        def t_UINT(t):
            r"\d+"
            t.value = int(t.value)
            return t

        def t_SIGN(t):
            r"[+-](?=\d)"
            t.value = int(t.value + "1")
            return t

        # This needs to be a function so we can force it to happen
        # before t_UNIT
        def t_FUNCNAME(t):
            r"((sqrt)|(ln)|(exp)|(log)|(mag)|(dB)|(dex))(?=\ *\()"
            return t

        # A possible unit is something that consists of characters not used
        # for anything else: no spaces, no digits, signs, periods, stars,
        # carets, parentheses or commas.
        def t_UNIT(t):
            r"[^\s\d+\-\./\*\^\(\)\,]+"
            t.value = cls._get_unit(t)
            return t

        t_ignore = " "

        # Error handling rule
        def t_error(t):
            raise ValueError(f"Invalid character at col {t.lexpos}")

        return parsing.lex(
            lextab="generic_lextab", package="astropy/units", reflags=int(re.UNICODE)
        )

    @classproperty(lazy=True)
    def _parser(cls) -> ThreadSafeParser:
        """
        The grammar here is based on the description in the `FITS
        standard
        <http://fits.gsfc.nasa.gov/standard30/fits_standard30aa.pdf>`_,
        Section 4.3, which is not terribly precise.  The exact grammar
        is here is based on the YACC grammar in the `unity library
        <https://bitbucket.org/nxg/unity/>`_.

        This same grammar is used by the `"fits"` and `"vounit"`
        formats, the only difference being the set of available unit
        strings.
        """
        tokens = cls._tokens

        def p_main(p):
            """
            main : unit
                 | structured_unit
                 | structured_subunit
            """
            if isinstance(p[1], tuple):
                # Unpack possible StructuredUnit inside a tuple, ie.,
                # ignore any set of very outer parentheses.
                p[0] = p[1][0]
            else:
                p[0] = p[1]

        def p_structured_subunit(p):
            """
            structured_subunit : OPEN_PAREN structured_unit CLOSE_PAREN
            """
            # We hide a structured unit enclosed by parentheses inside
            # a tuple, so that we can easily distinguish units like
            # "(au, au/day), yr" from "au, au/day, yr".
            p[0] = (p[2],)

        def p_structured_unit(p):
            """
            structured_unit : subunit COMMA
                            | subunit COMMA subunit
            """
            from astropy.units.structured import StructuredUnit

            inputs = (p[1],) if len(p) == 3 else (p[1], p[3])
            units = ()
            for subunit in inputs:
                if isinstance(subunit, tuple):
                    # Structured unit that should be its own entry in the
                    # new StructuredUnit (was enclosed in parentheses).
                    units += subunit
                elif isinstance(subunit, StructuredUnit):
                    # Structured unit whose entries should be
                    # individually added to the new StructuredUnit.
                    units += subunit.values()
                else:
                    # Regular unit to be added to the StructuredUnit.
                    units += (subunit,)

            p[0] = StructuredUnit(units)

        def p_subunit(p):
            """
            subunit : unit
                    | structured_unit
                    | structured_subunit
            """
            p[0] = p[1]

        def p_unit(p):
            """
            unit : product_of_units
                 | factor product_of_units
                 | factor PRODUCT product_of_units
                 | division_product_of_units
                 | factor division_product_of_units
                 | factor PRODUCT division_product_of_units
                 | inverse_unit
                 | factor inverse_unit
                 | factor PRODUCT inverse_unit
                 | factor
            """
            if len(p) == 2:
                p[0] = Unit(p[1])
            elif len(p) == 3:
                p[0] = CompositeUnit(p[1] * p[2].scale, p[2].bases, p[2].powers)
            elif len(p) == 4:
                p[0] = CompositeUnit(p[1] * p[3].scale, p[3].bases, p[3].powers)

        def p_division_product_of_units(p):
            """
            division_product_of_units : division_product_of_units DIVISION product_of_units
                                      | product_of_units
            """
            if len(p) == 4:
                p[0] = Unit(p[1] / p[3])
            else:
                p[0] = p[1]

        def p_inverse_unit(p):
            """
            inverse_unit : DIVISION unit_expression
            """
            p[0] = p[2] ** -1

        def p_factor(p):
            """
            factor : factor_fits
                   | factor_float
                   | factor_int
            """
            p[0] = p[1]

        def p_factor_float(p):
            """
            factor_float : signed_float
                         | signed_float UINT signed_int
                         | signed_float UINT POWER numeric_power
            """
            if cls.name == "fits":
                raise ValueError("Numeric factor not supported by FITS")
            if len(p) == 4:
                p[0] = p[1] * p[2] ** float(p[3])
            elif len(p) == 5:
                p[0] = p[1] * p[2] ** float(p[4])
            elif len(p) == 2:
                p[0] = p[1]

        def p_factor_int(p):
            """
            factor_int : UINT
                       | UINT signed_int
                       | UINT POWER numeric_power
                       | UINT UINT signed_int
                       | UINT UINT POWER numeric_power
            """
            if cls.name == "fits":
                raise ValueError("Numeric factor not supported by FITS")
            if len(p) == 2:
                p[0] = p[1]
            elif len(p) == 3:
                p[0] = p[1] ** float(p[2])
            elif len(p) == 4:
                if isinstance(p[2], int):
                    p[0] = p[1] * p[2] ** float(p[3])
                else:
                    p[0] = p[1] ** float(p[3])
            elif len(p) == 5:
                p[0] = p[1] * p[2] ** p[4]

        def p_factor_fits(p):
            """
            factor_fits : UINT POWER OPEN_PAREN signed_int CLOSE_PAREN
                        | UINT POWER OPEN_PAREN UINT CLOSE_PAREN
                        | UINT POWER signed_int
                        | UINT POWER UINT
                        | UINT SIGN UINT
                        | UINT OPEN_PAREN signed_int CLOSE_PAREN
            """
            if p[1] != 10:
                if cls.name == "fits":
                    raise ValueError("Base must be 10")
                else:
                    return
            if len(p) == 4:
                if p[2] in ("**", "^"):
                    p[0] = 10 ** p[3]
                else:
                    p[0] = 10 ** (p[2] * p[3])
            elif len(p) == 5:
                p[0] = 10 ** p[3]
            elif len(p) == 6:
                p[0] = 10 ** p[4]

        def p_product_of_units(p):
            """
            product_of_units : unit_expression PRODUCT product_of_units
                             | unit_expression product_of_units
                             | unit_expression
            """
            if len(p) == 2:
                p[0] = p[1]
            elif len(p) == 3:
                p[0] = p[1] * p[2]
            else:
                p[0] = p[1] * p[3]

        def p_unit_expression(p):
            """
            unit_expression : function
                            | unit_with_power
                            | OPEN_PAREN product_of_units CLOSE_PAREN
            """
            if len(p) == 2:
                p[0] = p[1]
            else:
                p[0] = p[2]

        def p_unit_with_power(p):
            """
            unit_with_power : UNIT POWER numeric_power
                            | UNIT numeric_power
                            | UNIT
            """
            if len(p) == 2:
                p[0] = p[1]
            elif len(p) == 3:
                p[0] = p[1] ** p[2]
            else:
                p[0] = p[1] ** p[3]

        def p_numeric_power(p):
            """
            numeric_power : sign UINT
                          | OPEN_PAREN paren_expr CLOSE_PAREN
            """
            if len(p) == 3:
                p[0] = p[1] * p[2]
            elif len(p) == 4:
                p[0] = p[2]

        def p_paren_expr(p):
            """
            paren_expr : sign UINT
                       | signed_float
                       | frac
            """
            if len(p) == 3:
                p[0] = p[1] * p[2]
            else:
                p[0] = p[1]

        def p_frac(p):
            """
            frac : sign UINT DIVISION sign UINT
            """
            p[0] = Fraction(p[1] * p[2], p[4] * p[5])

        def p_sign(p):
            """
            sign : SIGN
                 |
            """
            if len(p) == 2:
                p[0] = p[1]
            else:
                p[0] = 1

        def p_signed_int(p):
            """
            signed_int : SIGN UINT
            """
            p[0] = p[1] * p[2]

        def p_signed_float(p):
            """
            signed_float : sign UINT
                         | sign UFLOAT
            """
            p[0] = p[1] * p[2]

        def p_function(p):
            """
            function : FUNCNAME OPEN_PAREN main CLOSE_PAREN
            """
            if p[1] == "sqrt":
                p[0] = p[3] ** 0.5
                return
            elif p[1] in ("mag", "dB", "dex"):
                try:
                    function_unit = cls._validate_unit(p[1])
                except KeyError:
                    raise ValueError(cls._invalid_unit_error_message(p[1])) from None
                # In Generic, this is callable, but that does not have to
                # be the case in subclasses (e.g., in VOUnit it is not).
                if callable(function_unit):
                    p[0] = function_unit(p[3])
                    return

            raise ValueError(f"'{p[1]}' is not a recognized function")

        def p_error(p):
            raise ValueError()

        return parsing.yacc(tabmodule="generic_parsetab", package="astropy/units")


class Generic(Base, _GenericParserMixin):
    """
    A "generic" format.

    The syntax of the format is based directly on the FITS standard,
    but instead of only supporting the units that FITS knows about, it
    supports any unit available in the `astropy.units` namespace.
    """

    @classproperty
    def _units(cls) -> dict[str, UnitBase]:
        return get_current_unit_registry().registry

    @classmethod
    def _validate_unit(
        cls, s: str, deprecations: DeprecatedUnitAction = DeprecatedUnitAction.WARN
    ) -> UnitBase:
        if s in cls._unit_symbols:
            s = cls._unit_symbols[s]

        elif not s.isascii():
            if s[0].startswith("°"):
                s = "deg" if len(s) == 1 else "deg_" + s[1:]
            if len(s) > 1 and s[-1] in cls._unit_suffix_symbols:
                s = s[:-1] + cls._unit_suffix_symbols[s[-1]]
            elif s.endswith("R\N{INFINITY}"):
                s = s[:-2] + "Ry"

        return cls._units[s]

    @classmethod
    def _invalid_unit_error_message(cls, unit: str) -> str:
        return f"{unit} is not a valid unit. {cls._did_you_mean_units(unit)}"

    _unit_symbols: ClassVar[dict[str, str]] = {
        "%": "percent",
        "\N{PRIME}": "arcmin",
        "\N{DOUBLE PRIME}": "arcsec",
        "\N{MODIFIER LETTER SMALL H}": "hourangle",
        "e\N{SUPERSCRIPT MINUS}": "electron",
    }

    _unit_suffix_symbols: ClassVar[dict[str, str]] = {
        "\N{CIRCLED DOT OPERATOR}": "sun",
        "\N{SUN}": "sun",
        "\N{CIRCLED PLUS}": "earth",
        "\N{EARTH}": "earth",
        "\N{JUPITER}": "jupiter",
        "\N{LATIN SUBSCRIPT SMALL LETTER E}": "_e",
        "\N{LATIN SUBSCRIPT SMALL LETTER P}": "_p",
    }

    _translations: ClassVar[dict[int, str]] = str.maketrans({"\N{MINUS SIGN}": "-"})
    """Character translations that should be applied before parsing a string."""

    _superscripts: Final[str] = (
        "\N{SUPERSCRIPT MINUS}"
        "\N{SUPERSCRIPT PLUS SIGN}"
        "\N{SUPERSCRIPT ZERO}"
        "\N{SUPERSCRIPT ONE}"
        "\N{SUPERSCRIPT TWO}"
        "\N{SUPERSCRIPT THREE}"
        "\N{SUPERSCRIPT FOUR}"
        "\N{SUPERSCRIPT FIVE}"
        "\N{SUPERSCRIPT SIX}"
        "\N{SUPERSCRIPT SEVEN}"
        "\N{SUPERSCRIPT EIGHT}"
        "\N{SUPERSCRIPT NINE}"
    )

    _superscript_translations: ClassVar[dict[int, int]] = str.maketrans(
        _superscripts, "-+0123456789"
    )
    _regex_superscript: ClassVar[Pattern[str]] = re.compile(
        f"[{_superscripts}]?[{_superscripts[2:]}]+"
    )

    @classmethod
    def _convert_superscript(cls, m: Match[str]) -> str:
        return f"({m.group().translate(cls._superscript_translations)})"

    @classmethod
    def parse(cls, s: str, debug: bool = False) -> UnitBase:
        if not isinstance(s, str):
            s = s.decode("ascii")
        elif not s.isascii():
            # common normalization of unicode strings to avoid
            # having to deal with multiple representations of
            # the same character. This normalizes to "composed" form
            # and will e.g. convert OHM SIGN to GREEK CAPITAL LETTER OMEGA
            s = unicodedata.normalize("NFC", s)
            # Translate some basic unicode items that we'd like to support on
            # input but are not standard.
            s = s.translate(cls._translations)

            # TODO: might the below be better done in the parser/lexer?
            # Translate superscripts to parenthesized numbers; this ensures
            # that mixes of superscripts and regular numbers fail.
            s = cls._regex_superscript.sub(cls._convert_superscript, s)

        result = cls._do_parse(s, debug)

        # Check for excess solidi, but exclude fractional exponents (accepted)
        n_slashes = s.count("/")
        if n_slashes > 1 and (n_slashes - len(re.findall(r"\(\d+/\d+\)", s))) > 1:
            warnings.warn(
                f"'{s}' contains multiple slashes, which is "
                "discouraged by the FITS standard",
                UnitsWarning,
            )
        return result

    @classmethod
    def format_exponential_notation(
        cls, val: UnitScale | np.number, format_spec: str = "g"
    ) -> str:
        return format(val, format_spec)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This file contains routines to verify the correctness of UCD strings.
"""

# STDLIB
import re

# LOCAL
from astropy.utils import data

__all__ = ["check_ucd", "parse_ucd"]


class UCDWords:
    """
    Manages a list of acceptable UCD words.

    Works by reading in a data file exactly as provided by IVOA.  This
    file resides in data/ucd1p-words.txt.
    """

    def __init__(self):
        self._primary = set()
        self._secondary = set()
        self._descriptions = {}
        self._capitalization = {}

        with data.get_pkg_data_fileobj("data/ucd1p-words.txt", encoding="ascii") as fd:
            for line in fd.readlines():
                if line.startswith("#"):
                    continue

                type, name, descr = (x.strip() for x in line.split("|"))
                name_lower = name.lower()
                if type in "QPEVC":
                    self._primary.add(name_lower)
                if type in "QSEVC":
                    self._secondary.add(name_lower)
                self._descriptions[name_lower] = descr
                self._capitalization[name_lower] = name

    def is_primary(self, name):
        """
        Returns True if *name* is a valid primary name.
        """
        return name.lower() in self._primary

    def is_secondary(self, name):
        """
        Returns True if *name* is a valid secondary name.
        """
        return name.lower() in self._secondary

    def get_description(self, name):
        """
        Returns the official English description of the given UCD
        *name*.
        """
        return self._descriptions[name.lower()]

    def normalize_capitalization(self, name):
        """
        Returns the standard capitalization form of the given name.
        """
        return self._capitalization[name.lower()]


_ucd_singleton = None


def parse_ucd(ucd, check_controlled_vocabulary=False, has_colon=False):
    """
    Parse the UCD into its component parts.

    Parameters
    ----------
    ucd : str
        The UCD string

    check_controlled_vocabulary : bool, optional
        If `True`, then each word in the UCD will be verified against
        the UCD1+ controlled vocabulary, (as required by the VOTable
        specification version 1.2), otherwise not.

    has_colon : bool, optional
        If `True`, the UCD may contain a colon (as defined in earlier
        versions of the standard).

    Returns
    -------
    parts : list
        The result is a list of tuples of the form:

            (*namespace*, *word*)

        If no namespace was explicitly specified, *namespace* will be
        returned as ``'ivoa'`` (i.e., the default namespace).

    Raises
    ------
    ValueError
        if *ucd* is invalid
    """
    global _ucd_singleton
    if _ucd_singleton is None:
        _ucd_singleton = UCDWords()

    if has_colon:
        m = re.search(r"[^A-Za-z0-9_.:;\-]", ucd)
    else:
        m = re.search(r"[^A-Za-z0-9_.;\-]", ucd)
    if m is not None:
        raise ValueError(f"UCD has invalid character '{m.group(0)}' in '{ucd}'")

    word_component_re = r"[A-Za-z0-9][A-Za-z0-9\-_]*"
    word_re = rf"{word_component_re}(\.{word_component_re})*"

    parts = ucd.split(";")
    words = []
    for i, word in enumerate(parts):
        colon_count = word.count(":")
        if colon_count == 1:
            ns, word = word.split(":", 1)
            if not re.match(word_component_re, ns):
                raise ValueError(f"Invalid namespace '{ns}'")
            ns = ns.lower()
        elif colon_count > 1:
            raise ValueError(f"Too many colons in '{word}'")
        else:
            ns = "ivoa"

        if not re.match(word_re, word):
            raise ValueError(f"Invalid word '{word}'")

        if ns == "ivoa" and check_controlled_vocabulary:
            if i == 0:
                if not _ucd_singleton.is_primary(word):
                    if _ucd_singleton.is_secondary(word):
                        raise ValueError(
                            f"Secondary word '{word}' is not valid as a primary word"
                        )
                    else:
                        raise ValueError(f"Unknown word '{word}'")
            else:
                if not _ucd_singleton.is_secondary(word):
                    if _ucd_singleton.is_primary(word):
                        raise ValueError(
                            f"Primary word '{word}' is not valid as a secondary word"
                        )
                    else:
                        raise ValueError(f"Unknown word '{word}'")

        try:
            normalized_word = _ucd_singleton.normalize_capitalization(word)
        except KeyError:
            normalized_word = word
        words.append((ns, normalized_word))

    return words


def check_ucd(ucd, check_controlled_vocabulary=False, has_colon=False):
    """
    Returns False if *ucd* is not a valid `unified content descriptor`_.

    Parameters
    ----------
    ucd : str
        The UCD string

    check_controlled_vocabulary : bool, optional
        If `True`, then each word in the UCD will be verified against
        the UCD1+ controlled vocabulary, (as required by the VOTable
        specification version 1.2), otherwise not.

    has_colon : bool, optional
        If `True`, the UCD may contain a colon (as defined in earlier
        versions of the standard).

    Returns
    -------
    valid : bool
    """
    if ucd is None:
        return True

    try:
        parse_ucd(
            ucd,
            check_controlled_vocabulary=check_controlled_vocabulary,
            has_colon=has_colon,
        )
    except ValueError:
        return False
    return True



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""|Cosmology| <-> ECSV I/O, using |Cosmology.read| and |Cosmology.write|.

This module provides functions to write/read a |Cosmology| object to/from an ECSV file.
The functions are registered with ``readwrite_registry`` under the format name
"ascii.ecsv".

We assume the following setup:

    >>> from pathlib import Path
    >>> from tempfile import TemporaryDirectory
    >>> temp_dir = TemporaryDirectory()

To see reading a Cosmology from an ECSV file, we first write a Cosmology to an ECSV
file:

    >>> from astropy.cosmology import Cosmology, Planck18
    >>> file = Path(temp_dir.name) / "file.ecsv"
    >>> Planck18.write(file)

    >>> with open(file) as f: print(f.read())
    # %ECSV 1.0
    # ---
    # datatype:
    # - {name: name, datatype: string}
    # - {name: H0, unit: km / (Mpc s), datatype: float64, description: Hubble ...}
    ...
    # meta: !!omap
    # - {Oc0: 0.2607}
    ...
    # schema: astropy-2.0
    name H0 Om0 Tcmb0 Neff m_nu Ob0
    Planck18 67.66 0.30966 2.7255 3.046 [0.0,0.0,0.06] 0.04897
    <BLANKLINE>

Now we can read the Cosmology from the ECSV file, constructing a new cosmological
instance identical to the ``Planck18`` cosmology from which it was generated.

    >>> cosmo = Cosmology.read(file)
    >>> print(cosmo)
    FlatLambdaCDM(name="Planck18", H0=67.66 km / (Mpc s), Om0=0.30966,
                Tcmb0=2.7255 K, Neff=3.046, m_nu=[0. 0. 0.06] eV, Ob0=0.04897)
    >>> cosmo == Planck18
    True

If a file already exists, attempting to write will raise an error unless
``overwrite=True``.

    >>> Planck18.write(file, overwrite=True)

By default the cosmology class is written to the Table metadata. This can be changed to
a column of the table using the ``cosmology_in_meta`` keyword argument.

    >>> file = Path(temp_dir.name) / "file2.ecsv"
    >>> Planck18.write(file, cosmology_in_meta=False)
    >>> with open(file) as f: print(f.read())
    # %ECSV 1.0
    # ---
    # datatype:
    # - {name: cosmology, datatype: string}
    # - {name: name, datatype: string}
    ...
    # meta: !!omap
    # - {Oc0: 0.2607}
    ...
    # schema: astropy-2.0
    cosmology name H0 Om0 Tcmb0 Neff m_nu Ob0
    FlatLambdaCDM Planck18 67.66 0.30966 2.7255 3.046 [0.0,0.0,0.06] 0.04897
    <BLANKLINE>

The ``cosmology`` information (column or metadata) may be omitted if the cosmology class
(or its string name) is passed as the ``cosmology`` keyword argument to
|Cosmology.read|. Alternatively, specific cosmology classes can be used to parse the
data.

    >>> from astropy.cosmology import FlatLambdaCDM
    >>> print(FlatLambdaCDM.read(file))
    FlatLambdaCDM(name="Planck18", H0=67.66 km / (Mpc s), Om0=0.30966,
                    Tcmb0=2.7255 K, Neff=3.046, m_nu=[0. 0. 0.06] eV, Ob0=0.04897)

When using a specific cosmology class, the class' default parameter values are used to
fill in any missing information.

For files with multiple rows of cosmological parameters, the ``index`` argument is
needed to select the correct row. The index can be an integer for the row number or, if
the table is indexed by a column, the value of that column. If the table is not indexed
and ``index`` is a string, the "name" column is used as the indexing column.

Here is an example where ``index`` is needed and can be either an integer (for the row
number) or the name of one of the cosmologies, e.g. 'Planck15'.

    >>> from astropy.cosmology import Planck13, Planck15, Planck18
    >>> from astropy.table import vstack
    >>> cts = vstack([c.to_format("astropy.table")
    ...               for c in (Planck13, Planck15, Planck18)],
    ...              metadata_conflicts='silent')
    >>> file = Path(temp_dir.name) / "file3.ecsv"
    >>> cts.write(file)
    >>> with open(file) as f: print(f.read())
    # %ECSV 1.0
    # ---
    # datatype:
    # - {name: name, datatype: string}
    ...
    # meta: !!omap
    # - {Oc0: 0.2607}
    ...
    # schema: astropy-2.0
    name H0 Om0 Tcmb0 Neff m_nu Ob0
    Planck13 67.77 0.30712 2.7255 3.046 [0.0,0.0,0.06] 0.048252
    Planck15 67.74 0.3075 2.7255 3.046 [0.0,0.0,0.06] 0.0486
    Planck18 67.66 0.30966 2.7255 3.046 [0.0,0.0,0.06] 0.04897

    >>> cosmo = Cosmology.read(file, index="Planck15", format="ascii.ecsv")
    >>> cosmo == Planck15
    True

Fields of the table in the file can be renamed to match the
`~astropy.cosmology.Cosmology` class' signature using the ``rename`` argument. This is
useful when the files's column names do not match the class' parameter names.

    >>> file = Path(temp_dir.name) / "file4.ecsv"
    >>> Planck18.write(file, rename={"H0": "Hubble"})
    >>> with open(file) as f: print(f.read())
    # %ECSV 1.0
    # ---
    # datatype:
    # - {name: name, datatype: string}
    ...
    # meta: !!omap
    # - {Oc0: 0.2607}
    ...
    # schema: astropy-2.0
    name Hubble Om0 Tcmb0 Neff m_nu Ob0
    Planck18 67.66 0.30966 2.7255 3.046 [0.0,0.0,0.06] 0.04897

    >>> cosmo = Cosmology.read(file, rename={"Hubble": "H0"})
    >>> cosmo == Planck18
    True

By default :class:`~astropy.cosmology.Cosmology` instances are written using
`~astropy.table.QTable` as an intermediate representation (for details see
|Cosmology.to_format|, with ``format="astropy.table"``). The `~astropy.table.Table` type
can be changed using the ``cls`` keyword argument.

    >>> from astropy.table import Table
    >>> file = Path(temp_dir.name) / "file5.ecsv"
    >>> Planck18.write(file, cls=Table)

For most use cases, the default ``cls`` of :class:`~astropy.table.QTable` is recommended
and will be largely indistinguishable from other table types, as the ECSV format is
agnostic to the table type. An example of a difference that might necessitate using a
different table type is if a different ECSV schema is desired.

Additional keyword arguments are passed to ``QTable.read`` and ``QTable.write``.

.. testcleanup::

    >>> temp_dir.cleanup()
"""

from collections.abc import Mapping
from typing import Any, TypeVar

import astropy.units as u
from astropy.table import QTable, Table

# isort: split
import astropy.cosmology.units as cu
from astropy.cosmology._src.core import Cosmology
from astropy.cosmology._src.io.connect import readwrite_registry
from astropy.cosmology._src.typing import _CosmoT
from astropy.io.typing import PathLike, ReadableFileLike, WriteableFileLike

from .table import from_table, to_table

_TableT = TypeVar("_TableT", bound=Table)


def read_ecsv(
    filename: PathLike | ReadableFileLike[Table],
    index: int | str | None = None,
    *,
    move_to_meta: bool = False,
    cosmology: str | type[_CosmoT] | None = None,
    rename: Mapping[str, str] | None = None,
    **kwargs: Any,
) -> _CosmoT:
    r"""Read a `~astropy.cosmology.Cosmology` from an ECSV file.

    Parameters
    ----------
    filename : path-like or file-like
        From where to read the Cosmology.
    index : int, str, or None, optional
        Needed to select the row in tables with multiple rows. ``index`` can be an
        integer for the row number or, if the table is indexed by a column, the value of
        that column. If the table is not indexed and ``index`` is a string, the "name"
        column is used as the indexing column.

    move_to_meta : bool (optional, keyword-only)
        Whether to move keyword arguments that are not in the Cosmology class' signature
        to the Cosmology's metadata. This will only be applied if the Cosmology does NOT
        have a keyword-only argument (e.g. ``**kwargs``). Arguments moved to the
        metadata will be merged with existing metadata, preferring specified metadata in
        the case of a merge conflict (e.g. for ``Cosmology(meta={'key':10}, key=42)``,
        the ``Cosmology.meta`` will be ``{'key': 10}``).

    cosmology : str or type or None (optional, keyword-only)
        The cosmology class (or string name thereof) to use when constructing the
        cosmology instance. The class also provides default parameter values, filling in
        any non-mandatory arguments missing in 'table'.

    rename : dict or None (optional, keyword-only)
        A dictionary mapping column names in 'table' to fields of the
        `~astropy.cosmology.Cosmology` class.

    **kwargs
        Passed to ``QTable.read``

    Returns
    -------
    `~astropy.cosmology.Cosmology` subclass instance

    Examples
    --------
    We assume the following setup:

        >>> from pathlib import Path
        >>> from tempfile import TemporaryDirectory
        >>> temp_dir = TemporaryDirectory()

    To see reading a Cosmology from an ECSV file, we first write a Cosmology to an ECSV
    file:

        >>> from astropy.cosmology import Cosmology, Planck18
        >>> file = Path(temp_dir.name) / "file.ecsv"
        >>> Planck18.write(file)

        >>> with open(file) as f: print(f.read())
        # %ECSV 1.0
        # ---
        # datatype:
        # - {name: name, datatype: string}
        ...
        # meta: !!omap
        # - {Oc0: 0.2607}
        ...
        # schema: astropy-2.0
        name H0 Om0 Tcmb0 Neff m_nu Ob0
        Planck18 67.66 0.30966 2.7255 3.046 [0.0,0.0,0.06] 0.04897
        <BLANKLINE>

    Now we can read the Cosmology from the ECSV file, constructing a new cosmological
    instance identical to the ``Planck18`` cosmology from which it was generated.

        >>> cosmo = Cosmology.read(file)
        >>> print(cosmo)
        FlatLambdaCDM(name="Planck18", H0=67.66 km / (Mpc s), Om0=0.30966,
                    Tcmb0=2.7255 K, Neff=3.046, m_nu=[0. 0. 0.06] eV, Ob0=0.04897)
        >>> cosmo == Planck18
        True

    The ``cosmology`` information (column or metadata) may be omitted if the cosmology
    class (or its string name) is passed as the ``cosmology`` keyword argument to
    |Cosmology.read|. Alternatively, specific cosmology classes can be used to parse the
    data.

        >>> from astropy.cosmology import FlatLambdaCDM
        >>> print(FlatLambdaCDM.read(file))
        FlatLambdaCDM(name="Planck18", H0=67.66 km / (Mpc s), Om0=0.30966,
                      Tcmb0=2.7255 K, Neff=3.046, m_nu=[0. 0. 0.06] eV, Ob0=0.04897)

    When using a specific cosmology class, the class' default parameter values are used
    to fill in any missing information.

    For files with multiple rows of cosmological parameters, the ``index`` argument is
    needed to select the correct row. The index can be an integer for the row number or,
    if the table is indexed by a column, the value of that column. If the table is not
    indexed and ``index`` is a string, the "name" column is used as the indexing column.

    Here is an example where ``index`` is needed and can be either an integer (for the
    row number) or the name of one of the cosmologies, e.g. 'Planck15'.

        >>> from astropy.cosmology import Planck13, Planck15, Planck18
        >>> from astropy.table import vstack
        >>> cts = vstack([c.to_format("astropy.table")
        ...               for c in (Planck13, Planck15, Planck18)],
        ...              metadata_conflicts='silent')
        >>> file = Path(temp_dir.name) / "file2.ecsv"
        >>> cts.write(file)
        >>> with open(file) as f: print(f.read())
        # %ECSV 1.0
        # ---
        # datatype:
        # - {name: name, datatype: string}
        ...
        # meta: !!omap
        # - {Oc0: 0.2607}
        ...
        # schema: astropy-2.0
        name H0 Om0 Tcmb0 Neff m_nu Ob0
        Planck13 67.77 0.30712 2.7255 3.046 [0.0,0.0,0.06] 0.048252
        Planck15 67.74 0.3075 2.7255 3.046 [0.0,0.0,0.06] 0.0486
        Planck18 67.66 0.30966 2.7255 3.046 [0.0,0.0,0.06] 0.04897

        >>> cosmo = Cosmology.read(file, index="Planck15", format="ascii.ecsv")
        >>> cosmo == Planck15
        True

    Fields of the table in the file can be renamed to match the
    `~astropy.cosmology.Cosmology` class' signature using the ``rename`` argument. This
    is useful when the files's column names do not match the class' parameter names.
    For this example we need to make a new file with renamed columns:

        >>> file = Path(temp_dir.name) / "file3.ecsv"
        >>> renamed_table = Planck18.to_format("astropy.table", rename={"H0": "Hubble"})
        >>> renamed_table.write(file)
        >>> with open(file) as f: print(f.read())
        # %ECSV 1.0
        # ---
        # datatype:
        # - {name: name, datatype: string}
        ...
        # meta: !!omap
        # - {Oc0: 0.2607}
        ...
        # schema: astropy-2.0
        name Hubble Om0 Tcmb0 Neff m_nu Ob0
        Planck18 67.66 0.30966 2.7255 3.046 [0.0,0.0,0.06] 0.04897

    Now we can read the Cosmology from the ECSV file, with the required renaming:

        >>> cosmo = Cosmology.read(file, rename={"Hubble": "H0"})
        >>> cosmo == Planck18
        True

    Additional keyword arguments are passed to ``QTable.read``.

    .. testcleanup::

        >>> temp_dir.cleanup()
    """
    kwargs["format"] = "ascii.ecsv"
    with u.add_enabled_units(cu):
        table = QTable.read(filename, **kwargs)

    # build cosmology from table
    return from_table(
        table,
        index=index,
        move_to_meta=move_to_meta,
        cosmology=cosmology,
        rename=rename,
    )


def write_ecsv(
    cosmology: Cosmology,
    file: PathLike | WriteableFileLike[_TableT],
    *,
    overwrite: bool = False,
    cls: type[_TableT] = QTable,
    cosmology_in_meta: bool = True,
    rename: Mapping[str, str] | None = None,
    **kwargs: Any,
) -> None:
    """Serialize the cosmology into a ECSV.

    Parameters
    ----------
    cosmology : `~astropy.cosmology.Cosmology`
        The cosmology instance to convert to a mapping.
    file : path-like or file-like
        Location to save the serialized cosmology.

    overwrite : bool (optional, keyword-only)
        Whether to overwrite the file, if it exists.
    cls : type (optional, keyword-only)
        Astropy :class:`~astropy.table.Table` (sub)class to use when writing. Default is
        :class:`~astropy.table.QTable`.
    cosmology_in_meta : bool (optional, keyword-only)
        Whether to put the cosmology class in the Table metadata (if `True`, default) or
        as the first column (if `False`).
    rename : Mapping[str, str] or None (optional keyword-only)
        A mapping of field names on the `~astropy.cosmology.Cosmology` to column names
        of the table.

    **kwargs
        Passed to ``cls.write``

    Raises
    ------
    TypeError
        If kwarg (optional) 'cls' is not a subclass of `astropy.table.Table`

    Examples
    --------
    We assume the following setup:

        >>> from pathlib import Path
        >>> from tempfile import TemporaryDirectory
        >>> temp_dir = TemporaryDirectory()

    A Cosmology can be written to an ECSV file:

        >>> from astropy.cosmology import Cosmology, Planck18
        >>> file = Path(temp_dir.name) / "file.ecsv"
        >>> Planck18.write(file)

        >>> with open(file) as f: print(f.read())
        # %ECSV 1.0
        # ---
        # datatype:
        # - {name: name, datatype: string}
        ...
        # meta: !!omap
        # - {Oc0: 0.2607}
        ...
        # schema: astropy-2.0
        name H0 Om0 Tcmb0 Neff m_nu Ob0
        Planck18 67.66 0.30966 2.7255 3.046 [0.0,0.0,0.06] 0.04897
        <BLANKLINE>

    If a file already exists, attempting to write will raise an error unless
    ``overwrite=True``.

        >>> Planck18.write(file, overwrite=True)

    By default :class:`~astropy.cosmology.Cosmology` instances are written using
    `~astropy.table.QTable` as an intermediate representation (for details see
    |Cosmology.to_format|, with ``format="astropy.table"``). The `~astropy.table.Table`
    type can be changed using the ``cls`` keyword argument.

        >>> from astropy.table import Table
        >>> file = Path(temp_dir.name) / "file2.ecsv"
        >>> Planck18.write(file, cls=Table)

    For most use cases, the default ``cls`` of :class:`~astropy.table.QTable` is
    recommended and will be largely indistinguishable from other table types, as the
    ECSV format is agnostic to the table type. An example of a difference that might
    necessitate using a different table type is if a different ECSV schema is desired.

    By default the cosmology class is written to the Table metadata. This can be changed
    to a column of the table using the ``cosmology_in_meta`` keyword argument.

        >>> file = Path(temp_dir.name) / "file3.ecsv"
        >>> Planck18.write(file, cosmology_in_meta=False)
        >>> with open(file) as f: print(f.read())
        # %ECSV 1.0
        # ---
        # datatype:
        # - {name: cosmology, datatype: string}
        # - {name: name, datatype: string}
        ...
        # meta: !!omap
        # - {Oc0: 0.2607}
        ...
        # schema: astropy-2.0
        cosmology name H0 Om0 Tcmb0 Neff m_nu Ob0
        FlatLambdaCDM Planck18 67.66 0.30966 2.7255 3.046 [0.0,0.0,0.06] 0.04897
        <BLANKLINE>

    Fields of the Cosmology can be renamed to when writing to an ECSV file using the
    ``rename`` argument.

        >>> file = Path(temp_dir.name) / "file4.ecsv"
        >>> Planck18.write(file, rename={"H0": "Hubble"})
        >>> with open(file) as f: print(f.read())
        # %ECSV 1.0
        # ---
        # datatype:
        # - {name: name, datatype: string}
        ...
        # meta:
        ...
        # schema: astropy-2.0
        name Hubble Om0 Tcmb0 Neff m_nu Ob0
        Planck18 67.66 0.30966 2.7255 3.046 [0.0,0.0,0.06] 0.04897
        <BLANKLINE>

    Additional keyword arguments are passed to :attr:`astropy.table.QTable.write`.

    .. testcleanup::

        >>> temp_dir.cleanup()
    """
    table = to_table(
        cosmology, cls=cls, cosmology_in_meta=cosmology_in_meta, rename=rename
    )

    kwargs["format"] = "ascii.ecsv"
    table.write(file, overwrite=overwrite, **kwargs)


def ecsv_identify(
    origin: object, filepath: str | None, *args: object, **kwargs: object
) -> bool:
    """Identify if object uses the Table format.

    Returns
    -------
    bool
    """
    return filepath is not None and filepath.endswith(".ecsv")


# ===================================================================
# Register

readwrite_registry.register_reader("ascii.ecsv", Cosmology, read_ecsv)
readwrite_registry.register_writer("ascii.ecsv", Cosmology, write_ecsv)
readwrite_registry.register_identifier("ascii.ecsv", Cosmology, ecsv_identify)



# Licensed under a 3-clause BSD style license - see LICENSE.rst


import numpy as np
import pytest
from hypothesis import given
from hypothesis.extra.numpy import arrays
from hypothesis.strategies import integers

from astropy import coordinates, time
from astropy import units as u
from astropy.table import Column, NdarrayMixin, QTable, Table, table_helpers, unique
from astropy.time import Time
from astropy.utils.exceptions import AstropyUserWarning


def sort_eq(list1, list2):
    return sorted(list1) == sorted(list2)


def test_column_group_by(T1q):
    """Test grouping a Column by various key types."""
    # T1q["a"] could be Column or Quantity, so force the object we want to group to be
    # Column. Then later we are using the "a" column as a grouping key.
    t1a = Column(T1q["a"])
    unit = T1q["a"].unit or 1

    # Group by a Column (i.e. numpy array)
    t1ag = t1a.group_by(T1q["a"])
    keys = t1ag.groups.keys
    assert np.all(t1ag.groups.indices == np.array([0, 1, 4, 8]))
    assert np.all(keys == np.array([0, 1, 2]) * unit)

    # Group by a Table and numpy structured array
    for t1ag, key_unit in (
        (t1a.group_by(T1q["a", "b"]), unit),
        (t1a.group_by(T1q["a", "b"].as_array()), 1),
    ):
        assert np.all(t1ag.groups.indices == np.array([0, 1, 3, 4, 5, 7, 8]))
        keys = t1ag.groups.keys
        assert keys.dtype.names == ("a", "b")
        assert np.all(keys["a"] == np.array([0, 1, 1, 2, 2, 2]) * key_unit)
        assert np.all(keys["b"] == np.array(["a", "a", "b", "a", "b", "c"]))


def test_column_group_by_no_argsort(T1b):
    t1a = T1b["a"]
    with pytest.raises(
        TypeError, match=r"keys input \(list\) must have an `argsort` method"
    ):
        # Pass a Python list with no argsort method
        t1a.group_by(list(range(len(t1a))))


def test_table_group_by(T1):
    """
    Test basic table group_by functionality for possible key types and for
    masked/unmasked tables.
    """
    for masked in (False, True):
        t1 = QTable(T1, masked=masked)
        # Group by a single column key specified by name
        tg = t1.group_by("a")
        assert np.all(tg.groups.indices == np.array([0, 1, 4, 8]))
        assert str(tg.groups) == "<TableGroups indices=[0 1 4 8]>"
        assert str(tg["a"].groups) == "<ColumnGroups indices=[0 1 4 8]>"

        # Sorted by 'a' and in original order for rest
        assert tg.pformat() == [
            " a   b   c   d   q ",
            "                 m ",
            "--- --- --- --- ---",
            "  0   a 0.0   4 4.0",
            "  1   b 3.0   5 5.0",
            "  1   a 2.0   6 6.0",
            "  1   a 1.0   7 7.0",
            "  2   c 7.0   0 0.0",
            "  2   b 5.0   1 1.0",
            "  2   b 6.0   2 2.0",
            "  2   a 4.0   3 3.0",
        ]
        assert tg.meta["ta"] == 1
        assert tg["c"].meta["a"] == 1
        assert tg["c"].description == "column c"

        # Group by a table column
        tg2 = t1.group_by(t1["a"])
        assert tg.pformat() == tg2.pformat()

        # Group by two columns spec'd by name
        for keys in (["a", "b"], ("a", "b")):
            tg = t1.group_by(keys)
            assert np.all(tg.groups.indices == np.array([0, 1, 3, 4, 5, 7, 8]))
            # Sorted by 'a', 'b' and in original order for rest
            assert tg.pformat() == [
                " a   b   c   d   q ",
                "                 m ",
                "--- --- --- --- ---",
                "  0   a 0.0   4 4.0",
                "  1   a 2.0   6 6.0",
                "  1   a 1.0   7 7.0",
                "  1   b 3.0   5 5.0",
                "  2   a 4.0   3 3.0",
                "  2   b 5.0   1 1.0",
                "  2   b 6.0   2 2.0",
                "  2   c 7.0   0 0.0",
            ]

        # Group by a Table
        tg2 = t1.group_by(t1["a", "b"])
        assert tg.pformat() == tg2.pformat()

        # Group by a structured array
        tg2 = t1.group_by(t1["a", "b"].as_array())
        assert tg.pformat() == tg2.pformat()

        # Group by a simple ndarray
        tg = t1.group_by(np.array([0, 1, 0, 1, 2, 1, 0, 0]))
        assert np.all(tg.groups.indices == np.array([0, 4, 7, 8]))
        assert tg.pformat() == [
            " a   b   c   d   q ",
            "                 m ",
            "--- --- --- --- ---",
            "  2   c 7.0   0 0.0",
            "  2   b 6.0   2 2.0",
            "  1   a 2.0   6 6.0",
            "  1   a 1.0   7 7.0",
            "  2   b 5.0   1 1.0",
            "  2   a 4.0   3 3.0",
            "  1   b 3.0   5 5.0",
            "  0   a 0.0   4 4.0",
        ]


def test_groups_keys(T1m: QTable):
    tg = T1m.group_by("a")
    unit = T1m["a"].unit or 1
    keys = tg.groups.keys
    assert keys.dtype.names == ("a",)
    assert np.all(keys["a"] == np.array([0, 1, 2]) * unit)

    tg = T1m.group_by(["a", "b"])
    keys = tg.groups.keys
    assert keys.dtype.names == ("a", "b")
    assert np.all(keys["a"] == np.array([0, 1, 1, 2, 2, 2]) * unit)
    assert np.all(keys["b"] == np.array(["a", "a", "b", "a", "b", "c"]))

    # Grouping by Column ignores column name
    tg = T1m.group_by(T1m["b"])
    keys = tg.groups.keys
    assert keys.dtype.names is None


def test_groups_keys_time(T1b: QTable):
    """Group a table with a time column using that column as a key."""
    T1b = T1b.copy()
    T1b["a"] = Time(T1b["a"], format="cxcsec")

    tg = T1b.group_by("a")
    keys = tg.groups.keys
    assert keys.dtype.names == ("a",)
    assert np.all(keys["a"] == Time(np.array([0, 1, 2]), format="cxcsec"))

    tg = T1b.group_by(["a", "b"])
    keys = tg.groups.keys
    assert keys.dtype.names == ("a", "b")
    assert np.all(keys["a"] == Time(np.array([0, 1, 1, 2, 2, 2]), format="cxcsec"))
    assert np.all(keys["b"] == np.array(["a", "a", "b", "a", "b", "c"]))


def test_groups_iterator(T1):
    tg = T1.group_by("a")
    for ii, group in enumerate(tg.groups):
        assert group.pformat() == tg.groups[ii].pformat()
        assert group["a"][0] == tg["a"][tg.groups.indices[ii]]


def test_grouped_copy(T1):
    """
    Test that copying a table or column copies the groups properly
    """
    for masked in (False, True):
        t1 = QTable(T1, masked=masked)
        tg = t1.group_by("a")
        tgc = tg.copy()
        assert np.all(tgc.groups.indices == tg.groups.indices)
        assert np.all(tgc.groups.keys == tg.groups.keys)

        tac = tg["a"].copy()
        assert np.all(tac.groups.indices == tg["a"].groups.indices)

        c1 = t1["a"].copy()
        gc1 = c1.group_by(t1["a"])
        gc1c = gc1.copy()
        assert np.all(gc1c.groups.indices == np.array([0, 1, 4, 8]))


def test_grouped_slicing(T1):
    """
    Test that slicing a table removes previous grouping
    """

    for masked in (False, True):
        t1 = QTable(T1, masked=masked)

        # Regular slice of a table
        tg = t1.group_by("a")
        tg2 = tg[3:5]
        assert np.all(tg2.groups.indices == np.array([0, len(tg2)]))
        assert tg2.groups.keys is None


def test_group_column_from_table(T1):
    """
    Group a column that is part of a table
    """
    cg = T1["c"].group_by(np.array(T1["a"]))
    assert np.all(cg.groups.keys == np.array([0, 1, 2]))
    assert np.all(cg.groups.indices == np.array([0, 1, 4, 8]))


def test_table_groups_mask_index(T1):
    """
    Use boolean mask as item in __getitem__ for groups
    """
    for masked in (False, True):
        t1 = Table(T1, masked=masked).group_by("a")

        t2 = t1.groups[np.array([True, False, True])]
        assert len(t2.groups) == 2
        assert t2.groups[0].pformat() == t1.groups[0].pformat()
        assert t2.groups[1].pformat() == t1.groups[2].pformat()
        assert np.all(t2.groups.keys["a"] == np.array([0, 2]))


def test_table_groups_array_index(T1):
    """
    Use numpy array as item in __getitem__ for groups
    """
    for masked in (False, True):
        t1 = Table(T1, masked=masked).group_by("a")

        t2 = t1.groups[np.array([0, 2])]
        assert len(t2.groups) == 2
        assert t2.groups[0].pformat() == t1.groups[0].pformat()
        assert t2.groups[1].pformat() == t1.groups[2].pformat()
        assert np.all(t2.groups.keys["a"] == np.array([0, 2]))


def test_table_groups_slicing(T1):
    """
    Test that slicing table groups works
    """

    for masked in (False, True):
        t1 = Table(T1, masked=masked).group_by("a")

        # slice(0, 2)
        t2 = t1.groups[0:2]
        assert len(t2.groups) == 2
        assert t2.groups[0].pformat() == t1.groups[0].pformat()
        assert t2.groups[1].pformat() == t1.groups[1].pformat()
        assert np.all(t2.groups.keys["a"] == np.array([0, 1]))

        # slice(1, 2)
        t2 = t1.groups[1:2]
        assert len(t2.groups) == 1
        assert t2.groups[0].pformat() == t1.groups[1].pformat()
        assert np.all(t2.groups.keys["a"] == np.array([1]))

        # slice(0, 3, 2)
        t2 = t1.groups[0:3:2]
        assert len(t2.groups) == 2
        assert t2.groups[0].pformat() == t1.groups[0].pformat()
        assert t2.groups[1].pformat() == t1.groups[2].pformat()
        assert np.all(t2.groups.keys["a"] == np.array([0, 2]))


def test_grouped_item_access(T1):
    """
    Test that column slicing preserves grouping
    """
    for masked in (False, True):
        t1 = Table(T1, masked=masked)

        # Regular slice of a table
        tg = t1.group_by("a")
        tgs = tg["a", "c", "d"]
        assert np.all(tgs.groups.keys == tg.groups.keys)
        assert np.all(tgs.groups.indices == tg.groups.indices)
        tgsa = tgs.groups.aggregate(np.sum)
        assert tgsa.pformat() == [
            " a   c    d ",
            "--- ---- ---",
            "  0  0.0   4",
            "  1  6.0  18",
            "  2 22.0   6",
        ]

        tgs = tg["c", "d"]
        assert np.all(tgs.groups.keys == tg.groups.keys)
        assert np.all(tgs.groups.indices == tg.groups.indices)
        tgsa = tgs.groups.aggregate(np.sum)
        assert tgsa.pformat() == [
            " c    d ",
            "---- ---",
            " 0.0   4",
            " 6.0  18",
            "22.0   6",
        ]


def test_mutable_operations(T1):
    """
    Operations like adding or deleting a row should removing grouping,
    but adding or removing or renaming a column should retain grouping.
    """
    for masked in (False, True):
        t1 = QTable(T1, masked=masked)

        # add row
        tg = t1.group_by("a")
        tg.add_row((0, "a", 3.0, 4, 4 * u.m))
        assert np.all(tg.groups.indices == np.array([0, len(tg)]))
        assert tg.groups.keys is None

        # remove row
        tg = t1.group_by("a")
        tg.remove_row(4)
        assert np.all(tg.groups.indices == np.array([0, len(tg)]))
        assert tg.groups.keys is None

        # add column
        tg = t1.group_by("a")
        indices = tg.groups.indices.copy()
        tg.add_column(Column(name="e", data=np.arange(len(tg))))
        assert np.all(tg.groups.indices == indices)
        assert np.all(tg["e"].groups.indices == indices)
        assert np.all(tg["e"].groups.keys == tg.groups.keys)

        # remove column (not key column)
        tg = t1.group_by("a")
        tg.remove_column("b")
        assert np.all(tg.groups.indices == indices)
        # Still has original key col names
        assert tg.groups.keys.dtype.names == ("a",)
        assert np.all(tg["a"].groups.indices == indices)

        # remove key column
        tg = t1.group_by("a")
        tg.remove_column("a")
        assert np.all(tg.groups.indices == indices)
        assert tg.groups.keys.dtype.names == ("a",)
        assert np.all(tg["b"].groups.indices == indices)

        # rename key column
        tg = t1.group_by("a")
        tg.rename_column("a", "aa")
        assert np.all(tg.groups.indices == indices)
        assert tg.groups.keys.dtype.names == ("a",)
        assert np.all(tg["aa"].groups.indices == indices)


def test_group_by_masked(T1):
    t1m = QTable(T1, masked=True)
    t1m["c"].mask[4] = True
    t1m["d"].mask[5] = True
    assert t1m.group_by("a").pformat() == [
        " a   b   c   d   q ",
        "                 m ",
        "--- --- --- --- ---",
        "  0   a  --   4 4.0",
        "  1   b 3.0  -- 5.0",
        "  1   a 2.0   6 6.0",
        "  1   a 1.0   7 7.0",
        "  2   c 7.0   0 0.0",
        "  2   b 5.0   1 1.0",
        "  2   b 6.0   2 2.0",
        "  2   a 4.0   3 3.0",
    ]


def test_group_by_errors(T1):
    """
    Appropriate errors get raised.
    """
    # Bad column name as string
    with pytest.raises(ValueError):
        T1.group_by("f")

    # Bad column names in list
    with pytest.raises(ValueError):
        T1.group_by(["f", "g"])

    # Wrong length array
    with pytest.raises(ValueError):
        T1.group_by(np.array([1, 2]))

    # Wrong type
    with pytest.raises(TypeError):
        T1.group_by(None)

    # Masked key column
    t1 = QTable(T1, masked=True)
    t1["a"].mask[4] = True
    with pytest.raises(ValueError):
        t1.group_by("a")


def test_groups_keys_meta(T1):
    """
    Make sure the keys meta['grouped_by_table_cols'] is working.
    """
    # Group by column in this table
    tg = T1.group_by("a")
    assert tg.groups.keys.meta["grouped_by_table_cols"] is True
    assert tg["c"].groups.keys.meta["grouped_by_table_cols"] is True
    assert tg.groups[1].groups.keys.meta["grouped_by_table_cols"] is True
    assert (
        tg["d"]
        .groups[np.array([False, True, True])]
        .groups.keys.meta["grouped_by_table_cols"]
        is True
    )

    # Group by external Table
    tg = T1.group_by(T1["a", "b"])
    assert tg.groups.keys.meta["grouped_by_table_cols"] is False
    assert tg["c"].groups.keys.meta["grouped_by_table_cols"] is False
    assert tg.groups[1].groups.keys.meta["grouped_by_table_cols"] is False

    # Group by external numpy array
    tg = T1.group_by(T1["a", "b"].as_array())
    assert not hasattr(tg.groups.keys, "meta")
    assert not hasattr(tg["c"].groups.keys, "meta")

    # Group by Column
    tg = T1.group_by(T1["a"])
    assert "grouped_by_table_cols" not in tg.groups.keys.meta
    assert "grouped_by_table_cols" not in tg["c"].groups.keys.meta


def test_table_aggregate(T1):
    """
    Aggregate a table
    """
    # Table with only summable cols
    t1 = T1["a", "c", "d"]
    tg = t1.group_by("a")
    tga = tg.groups.aggregate(np.sum)
    assert tga.pformat() == [
        " a   c    d ",
        "--- ---- ---",
        "  0  0.0   4",
        "  1  6.0  18",
        "  2 22.0   6",
    ]
    # Reverts to default groups
    assert np.all(tga.groups.indices == np.array([0, 3]))
    assert tga.groups.keys is None

    # metadata survives
    assert tga.meta["ta"] == 1
    assert tga["c"].meta["a"] == 1
    assert tga["c"].description == "column c"

    # Aggregate with np.sum with masked elements.  This results
    # in one group with no elements, hence a nan result and conversion
    # to float for the 'd' column.
    t1m = QTable(T1, masked=True)
    t1m["c"].mask[4:6] = True
    t1m["d"].mask[4:6] = True
    t1m["q"].mask[4:6] = True
    tg = t1m.group_by("a")

    with (
        pytest.warns(UserWarning, match="converting a masked element to nan"),
        pytest.warns(AstropyUserWarning, match="Cannot aggregate column"),
    ):
        tga = tg.groups.aggregate(np.sum)

    assert tga.pformat() == [
        " a   c    d    q  ",
        "               m  ",
        "--- ---- ---- ----",
        "  0   --   --  ———",
        "  1  3.0 13.0  ———",
        "  2 22.0  6.0  6.0",
    ]

    # Aggregate with np.sum with masked elements, but where every
    # group has at least one remaining (unmasked) element.  Then
    # the int column stays as an int.
    t1m = QTable(t1, masked=True)
    t1m["c"].mask[5] = True
    t1m["d"].mask[5] = True
    tg = t1m.group_by("a")
    tga = tg.groups.aggregate(np.sum)
    assert tga.pformat() == [
        " a   c    d ",
        "--- ---- ---",
        "  0  0.0   4",
        "  1  3.0  13",
        "  2 22.0   6",
    ]

    # Aggregate with a column type that cannot by supplied to the aggregating
    # function.  This raises a warning but still works.
    tg = T1.group_by("a")
    with pytest.warns(AstropyUserWarning, match="Cannot aggregate column"):
        tga = tg.groups.aggregate(np.sum)
    assert tga.pformat() == [
        " a   c    d   q  ",
        "              m  ",
        "--- ---- --- ----",
        "  0  0.0   4  4.0",
        "  1  6.0  18 18.0",
        "  2 22.0   6  6.0",
    ]


def test_table_aggregate_reduceat(T1):
    """
    Aggregate table with functions which have a reduceat method
    """

    # Comparison functions without reduceat
    def np_mean(x):
        return np.mean(x)

    def np_sum(x):
        return np.sum(x)

    def np_add(x):
        return np.add(x)

    # Table with only summable cols
    t1 = T1["a", "c", "d"]
    tg = t1.group_by("a")
    # Comparison
    tga_r = tg.groups.aggregate(np.sum)
    tga_a = tg.groups.aggregate(np.add)
    tga_n = tg.groups.aggregate(np_sum)

    assert np.all(tga_r == tga_n)
    assert np.all(tga_a == tga_n)
    assert tga_n.pformat() == [
        " a   c    d ",
        "--- ---- ---",
        "  0  0.0   4",
        "  1  6.0  18",
        "  2 22.0   6",
    ]

    tga_r = tg.groups.aggregate(np.mean)
    tga_n = tg.groups.aggregate(np_mean)
    assert np.all(tga_r == tga_n)
    assert tga_n.pformat() == [
        " a   c   d ",
        "--- --- ---",
        "  0 0.0 4.0",
        "  1 2.0 6.0",
        "  2 5.5 1.5",
    ]

    # Binary ufunc np_add should raise warning without reduceat
    t2 = T1["a", "c"]
    tg = t2.group_by("a")

    with pytest.warns(AstropyUserWarning, match="Cannot aggregate column"):
        tga = tg.groups.aggregate(np_add)
    assert tga.pformat() == [" a ", "---", "  0", "  1", "  2"]


def test_table_aggregate_reduceat_empty():
    for masked in (False, True):
        tg = Table(
            {
                "action": np.asarray([], dtype=str),
                "duration": np.asarray([], dtype=float),
            },
            masked=masked,
        )
        tga = tg.group_by("action").groups.aggregate(np.sum)
        assert tga.pformat() == ["action duration", "------ --------"]


def test_column_aggregate(T1):
    """
    Aggregate a single table column
    """
    for masked in (False, True):
        tg = QTable(T1, masked=masked).group_by("a")
        tga = tg["c"].groups.aggregate(np.sum)
        assert tga.pformat() == [" c  ", "----", " 0.0", " 6.0", "22.0"]


def test_column_aggregate_f8():
    """https://github.com/astropy/astropy/issues/12706"""
    # Just want to make sure it does not crash again.
    for masked in (False, True):
        tg = Table({"a": np.arange(2, dtype=">f8")}, masked=masked).group_by("a")
        tga = tg["a"].groups.aggregate(np.sum)
        assert tga.pformat() == [" a ", "---", "0.0", "1.0"]


def test_table_filter():
    """
    Table groups filtering
    """

    def all_positive(table, key_colnames):
        return all(
            np.all(table[colname] >= 0)
            for colname in table.colnames
            if colname not in key_colnames
        )

    # Negative value in 'a' column should not filter because it is a key col
    t = Table.read(
        [
            " a c d",
            " -2 7.0 0",
            " -2 5.0 1",
            " 0 0.0 4",
            " 1 3.0 5",
            " 1 2.0 -6",
            " 1 1.0 7",
            " 3 3.0 5",
            " 3 -2.0 6",
            " 3 1.0 7",
        ],
        format="ascii",
    )
    tg = t.group_by("a")
    t2 = tg.groups.filter(all_positive)
    assert t2.groups[0].pformat() == [
        " a   c   d ",
        "--- --- ---",
        " -2 7.0   0",
        " -2 5.0   1",
    ]
    assert t2.groups[1].pformat() == [" a   c   d ", "--- --- ---", "  0 0.0   4"]


def test_column_filter():
    """
    Table groups filtering
    """

    # Negative value in 'a' column should not filter because it is a key col
    t = Table.read(
        [
            " a c d",
            " -2 7.0 0",
            " -2 5.0 1",
            " 0 0.0 4",
            " 1 3.0 5",
            " 1 2.0 -6",
            " 1 1.0 7",
            " 3 3.0 5",
            " 3 -2.0 6",
            " 3 1.0 7",
        ],
        format="ascii",
    )
    tg = t.group_by("a")
    c2 = tg["c"].groups.filter(lambda column: np.all(column >= 0))
    assert len(c2.groups) == 3
    assert c2.groups[0].pformat() == [" c ", "---", "7.0", "5.0"]
    assert c2.groups[1].pformat() == [" c ", "---", "0.0"]
    assert c2.groups[2].pformat() == [" c ", "---", "3.0", "2.0", "1.0"]


def test_group_mixins():
    """
    Test grouping a table with mixin columns
    """
    # Setup mixins
    idx = np.arange(4)
    x = np.array([3.0, 1.0, 2.0, 1.0])
    q = x * u.m
    lon = coordinates.Longitude(x * u.deg)
    lat = coordinates.Latitude(x * u.deg)
    # For Time do J2000.0 + few * 0.1 ns (this requires > 64 bit precision)
    tm = time.Time(2000, format="jyear") + time.TimeDelta(x * 1e-10, format="sec")
    sc = coordinates.SkyCoord(ra=lon, dec=lat)
    aw = table_helpers.ArrayWrapper(x)
    nd = np.array([(3, "c"), (1, "a"), (2, "b"), (1, "a")], dtype="<i4,|S1").view(
        NdarrayMixin
    )

    qt = QTable(
        [idx, x, q, lon, lat, tm, sc, aw, nd],
        names=["idx", "x", "q", "lon", "lat", "tm", "sc", "aw", "nd"],
    )

    # Test group_by with each supported mixin type
    mixin_keys = ["x", "q", "lon", "lat", "tm", "sc", "aw", "nd"]
    for key in mixin_keys:
        qtg = qt.group_by(key)

        # Test that it got the sort order correct
        assert np.all(qtg["idx"] == [1, 3, 2, 0])

        # Test that the groups are right
        # Note: skip testing SkyCoord column because that doesn't have equality
        for name in ["x", "q", "lon", "lat", "tm", "aw", "nd"]:
            assert np.all(qt[name][[1, 3]] == qtg.groups[0][name])
            assert np.all(qt[name][[2]] == qtg.groups[1][name])
            assert np.all(qt[name][[0]] == qtg.groups[2][name])

    # Test that unique also works with mixins since most of the work is
    # done with group_by().  This is using *every* mixin as key.
    uqt = unique(qt, keys=mixin_keys)
    assert len(uqt) == 3
    assert np.all(uqt["idx"] == [1, 2, 0])
    assert np.all(uqt["x"] == [1.0, 2.0, 3.0])

    # Column group_by() with mixins
    idxg = qt["idx"].group_by(qt[mixin_keys])
    assert np.all(idxg == [1, 3, 2, 0])


@pytest.mark.parametrize(
    "col",
    [
        time.TimeDelta([1, 2], format="sec"),
        time.Time([1, 2], format="cxcsec"),
        coordinates.SkyCoord([1, 2], [3, 4], unit="deg,deg"),
    ],
)
def test_group_mixins_unsupported(col):
    """Test that aggregating unsupported mixins produces a warning only"""

    t = Table([[1, 1], [3, 4], col], names=["a", "b", "mix"])
    tg = t.group_by("a")

    with pytest.warns(AstropyUserWarning, match="Cannot aggregate column 'mix'"):
        tg.groups.aggregate(np.sum)


@pytest.mark.parametrize("add_index", [False, True])
@given(arrays("int64", shape=1000, elements=integers(min_value=0, max_value=5)))
def test_group_stable_sort(add_index, a):
    """Test that group_by preserves the order of the table.

    This table has 5 groups with an average of 200 rows per group, so it is not
    statistically possible that the groups will be in order by chance.

    This tests explicitly the case where grouping is done via the index sort.
    See: https://github.com/astropy/astropy/issues/14882
    """
    b = np.arange(len(a))
    t = Table([a, b], names=["a", "b"])
    if add_index:
        t.add_index("a")
    tg = t.group_by("a")
    for grp in tg.groups:
        assert np.all(grp["b"] == np.sort(grp["b"]))



"""
ECSV Engine Module
--------------------------
This module provides functionality for reading and writing Enhanced Character Separated
Values (ECSV) files using various backends, including:
- `PyArrow <https://arrow.apache.org/docs/python/>`_
- `pandas <https://pandas.pydata.org/>`_
- `Astropy's ASCII engine <https://docs.astropy.org/en/stable/io/ascii/>`_

ECSV is a human-readable, YAML-encoded table format used in the Astropy
ecosystem for storing tables with metadata, units, and complex data types.

Key Features
------------
- Defines data structures for representing ECSV column and header metadata.
- Implements multiple ECSV reader engines, supporting PyArrow, pandas, and Astropy's
  ASCII CSV readers.
- Handles conversion between ECSV datatypes and numpy/pandas/pyarrow types, including
  support for JSON-encoded columns, multidimensional arrays, and masked data.
- Provides robust parsing of ECSV headers and data, including support for compressed
  files and in-memory file-like objects.
- Ensures compatibility with legacy ECSV files and provides liberal error handling for
  unknown datatypes.
- Integrates with Astropy's Unified I/O registry for seamless reading and writing of
  ECSV files.

Main Classes and Functions
--------------------------
- ``ColumnECSV``: Represents the attributes of a column as described in the ECSV header.
- ``ECSVHeader``: Encapsulates the parsed ECSV header, including column definitions and
  table metadata.
- ``ECSVEngine`` and subclasses: Abstract base class and concrete implementations for
  different CSV parsing engines.
- ``read_ecsv``: Reads an ECSV file and returns an Astropy ``Table`` object, handling all
  necessary conversions and metadata.
- ``write_ecsv``: Writes an Astropy ``Table`` to an ECSV file.
- ``register_pyarrow_ecsv_table``: Registers the PyArrow ECSV reader/writer with
  Astropy's I/O registry.

Usage
-----
This module is intended for internal use within Astropy and for advanced users who need
fine-grained control over ECSV parsing and engine selection. For most users, reading and
writing ECSV files can be accomplished via the high-level ``Table.read`` and
``Table.write`` interfaces, e.g.:
```
Table.read(filename, format="ecsv", engine="pyarrow.csv")
```

Dependencies
------------
- numpy
- astropy.table
- pyarrow (optional, for PyArrow engine)
- pandas (optional, for pandas engine)
"""

import abc
import collections
import functools
import io
import json
import os
import re
import warnings
from collections.abc import Iterable
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Final, NamedTuple

import numpy as np
import numpy.typing as npt

from astropy.utils.data import get_readable_fileobj

if TYPE_CHECKING:
    from astropy.table import SerializedColumn, Table

__all__ = [
    "ColumnECSV",
    "ECSVEngine",
    "ECSVEngineIoAscii",
    "ECSVEnginePandas",
    "ECSVEnginePyArrow",
    "ECSVHeader",
    "read_ecsv",
    "register_ecsv_table",
    "write_ecsv",
]

ECSVEngines: Final[dict[str, "ECSVEngine"]] = {}


class DerivedColumnProperties(NamedTuple):
    """Named tuple for derived properties of a ECSV column specification.

    Attributes
    ----------
    csv_np_type : str
        Numpy type string for the CSV column data, e.g. "int64", "float32", "str".
        This is derived from the ECSV `datatype` and `subtype`.
    dtype : str
        Numpy dtype in the final column data. This may differ from `csv_np_type` in
        some cases, e.g. for JSON-encoded columns.
    shape : tuple[int, ...]
        Shape of the final column data as a tuple of integers. This is derived from the
        ECSV `subtype` if applicable, or an empty tuple for scalar columns.
    """

    csv_np_type: str
    dtype: str
    shape: tuple[int, ...]


@dataclass(frozen=True)
class ColumnECSV:
    """
    Class representing attributes of a column in an ECSV header.

    Attributes
    ----------
    name : str
        The name of the column.
    datatype : str
        The data type of the column as specified in the ECSV header.
    subtype : str or None, optional
        The subtype of the column, if applicable.
    unit : str or None, optional
        The unit of the column values, if specified.
    description : str or None, optional
        A description of the column.
    format : str or None, optional
        The format string for the column values.
    meta : dict or None, optional
        Additional metadata associated with the column.

    Properties
    ----------
    csv_np_type : str
        Numpy type string describing the column CSV data. In practice this is the same
        as the ECSV ``datatype`` except that "string" => "str". This is provided to the
        engine ``convert_np_type()`` method to generate the engine-specific type
        provided to the CSV reader. For instance, for pandas the ``int32`` type gets
        converted to ``Int32`` to read columns as a nullable int32.
    dtype : np.dtype
        Numpy dtype in the final column data. This may be entirely different from
        ``csv_np_type`` in some cases, in particular JSON-encoded fields.
    shape : tuple of int
        Shape of the final column data.
    """

    name: str
    datatype: str
    subtype: str | None = None
    unit: str | None = None
    description: str | None = None
    format: str | None = None
    meta: dict | None = None

    @functools.cached_property
    def csv_np_type(self) -> str:
        """Numpy type string describing the column CSV data."""
        return self._derived_properties.csv_np_type

    @functools.cached_property
    def dtype(self) -> np.dtype:
        """Numpy dtype in the final column data"""
        return np.dtype(self._derived_properties.dtype)

    @functools.cached_property
    def shape(self) -> tuple[int, ...]:
        """Shape of the column data"""
        return self._derived_properties.shape

    @functools.cached_property
    def _derived_properties(self) -> DerivedColumnProperties:
        """Get the csv_np_type, dtype, and shape of the column from ECSV header."""
        return get_csv_np_type_dtype_shape(self.datatype, self.subtype, self.name)


@dataclass(slots=True, frozen=True)
class ECSVHeader:
    """
    Class representing the information in an ECSV header.

    Attributes
    ----------
    n_header : int
        Total number of header lines in the ECSV file (including empty).
    n_empty : int
        Number of empty lines in the header section.
    n_comment : int
        Number of comment lines in the header section.
    cols : list[ColumnECSV]
        List of ``ColumnECSV`` objects describing the attributes of each column in
        header.
    table_meta : dict
        Metadata associated with the table, typically parsed from the ECSV header.
    delimiter : str
        Delimiter character used to separate columns in the ECSV file.
    """

    n_header: int
    n_empty: int
    n_comment: int
    cols: list[ColumnECSV]
    table_meta: dict
    delimiter: str


class ECSVEngine(metaclass=abc.ABCMeta):
    """Base class for ECSV reader engines.

    An engine is responsible for reading the raw CSV data that follows the ECSV header.
    This assumes that the engine has a defined Table Unified I/O interface.

    - `name` and `format` must be defined as class attributes in subclasses.
    - `engines` is a base class-level dictionary that maps engine names to their
      respective engine classes. Subclasses should not modify this directly.

    Properties
    ----------
    name : str
        Name of the engine, used for ``engine`` parameter in a call like:
        ``Table.read(filename, format="ecsv", engine="pyarrow")``.
    format : str
        Format string for the engine CSV reader, e.g. "pyarrow.csv", "ascii.csv", etc.
    engines : dict[str, ECSVEngine]
        Dictionary mapping engine names to their respective engine classes.
    """

    name: str | None = None
    format: str | None = None
    engines: dict[str, "ECSVEngine"] = {}

    def __init_subclass__(cls, **kwargs):
        """Register the subclass as an ECSV engine."""
        super().__init_subclass__(**kwargs)

        # Ensure that the subclass has the required string class attributes.
        for attr in ("name", "format"):
            if not isinstance(val := getattr(cls, attr, None), str):
                raise TypeError(
                    f"Subclasses of ECSVEngine must define a class attribute '{attr}' "
                    f"as a string, got {type(val)}."
                )

        cls.engines[cls.name] = cls

    @abc.abstractmethod
    def convert_np_type(self, np_type: str) -> Any:
        """
        Convert a numpy type string to engine-specific type for parsing.

        For instance, for pandas the ``"int32"`` numpy type gets converted to an
        ``Int32Dtype()`` instance to read columns as a nullable int32.

        Parameters
        ----------
        np_type : str
            The numpy type string to be converted.

        Returns
        -------
        Any
            Corresponding engine-specific type.
        """

    def get_converters(self, header):
        """
        Get a dictionary of converters for the columns in the ECSV header.

        This is used to convert column names to engine-specific types.

        Parameters
        ----------
        header : ECSVHeader
            The ECSV header containing column definitions.

        Returns
        -------
        dict[str, Any]
            Dictionary mapping column names to engine-specific converters.
        """
        return {col.name: self.convert_np_type(col.csv_np_type) for col in header.cols}

    @abc.abstractmethod
    def get_data_kwargs(
        self,
        header: ECSVHeader,
        null_values: list[str],
    ) -> dict[str, Any]:
        """
        Generate a dictionary of keyword arguments for data parsing.

        This accounts for the API variations in each engine CSV reader.

        Parameters
        ----------
        header : ECSVHeader
            ECSVHeader object within header information.
        null_values : list of str
            List of strings with values to be interpreted as null or missing data.

        Returns
        -------
        dict[str, Any]
            Dict of keyword arguments to be passed to engine CSV reader.
        """


class ECSVEnginePyArrow(ECSVEngine):
    """ECSV reader engine using PyArrow."""

    name = "pyarrow"
    format = "pyarrow.csv"

    def convert_np_type(self, np_type: str) -> str:
        # PyArrow does not support float128 and there is no workaround (unlike float16).
        if np_type == "float128":
            raise TypeError(
                "pyarrow engine does not support float128, choose a different engine"
            )

        # PyArrow does not support float16, so we need to convert it to float32.
        # The final output is still cast as float16.
        return "float32" if np_type == "float16" else np_type

    def get_data_kwargs(
        self,
        header: ECSVHeader,
        null_values: list[str],
    ) -> dict[str, Any]:
        # See base method for details.
        kw = {}
        kw["null_values"] = null_values
        kw["header_start"] = header.n_header
        kw["dtypes"] = self.get_converters(header)
        return kw


class ECSVEngineIoAscii(ECSVEngine):
    """ECSV reader engine using astropy.io.ascii Python CSV reader."""

    name = "io.ascii"
    format = "ascii.csv"

    def convert_np_type(self, np_type: str) -> np.generic:
        # Convert the np_type string to a numpy dtype type like np.int32, np.float64,
        # etc. This output is compatible with io.ascii `converters` option where is gets
        # used.
        return np.dtype(np_type).type

    def get_data_kwargs(
        self,
        header: ECSVHeader,
        null_values: list[str],
    ) -> dict[str, Any]:
        kw = {}
        kw["fill_values"] = get_null_values_per_column(
            header.cols, header.table_meta, null_values
        )
        kw["header_start"] = header.n_header - header.n_empty
        kw["converters"] = self.get_converters(header)
        # Fast reader does not support converters (defining types in advance) nor any
        # encoding. Converters are required, e.g. for a string column that looks like
        # floats. Would be nice to fix this, but in mean time use Python CSV reader.
        kw["fast_reader"] = False
        kw["strip_column_names"] = False
        return kw


class ECSVEnginePandas(ECSVEngine):
    """ECSV reader engine using pandas."""

    name = "pandas"
    format = "pandas.csv"

    def convert_np_type(self, np_type: str) -> np.dtype:
        # Convert the np_type to a pandas dtype will support for nullable types.
        import pandas as pd

        dtype = np.dtype(np_type)
        if dtype.kind in ("i", "u"):
            # Convert int64 to Int64, uint32 to UInt32, etc for nullable types
            converter = dtype.name.replace("i", "I").replace("u", "U")
        elif dtype.kind == "b":
            converter = "boolean"
        else:
            converter = np_type
        return pd.api.types.pandas_dtype(converter)

    def get_data_kwargs(
        self,
        header: ECSVHeader,
        null_values: list[str],
    ) -> dict[str, Any]:
        fill_values = get_null_values_per_column(
            header.cols, header.table_meta, null_values
        )
        null_values = collections.defaultdict(list)
        converters = self.get_converters(header)
        for null_value, _, col_name in fill_values:
            null_values[col_name].append(null_value)
            # Pandas parser does not natively parse nan or NaN for floats, so we need
            # to declare this as a null value.
            if converters[col_name].kind == "f":
                for nan in ("nan", "NaN"):
                    null_values[col_name].append(nan)

        kw = {
            "na_values": null_values,
            "keep_default_na": False,
            "comment": "#",
            "dtype": converters,
        }
        # Would prefer setting `"skiprows": header.n_header` above (as in the original
        # implementation prior to #18756) instead of "comment": "#". However there is a
        # bug in pandas.read_csv where skiprows does not work when the line includes a
        # quote character, see https://github.com/pandas-dev/pandas/issues/62739.
        return kw  # noqa: RET504


def is_numpy_dtype(np_type: str) -> bool:
    # Check if the given dtype is a valid numpy dtype.
    try:
        np.dtype(np_type)
    except Exception:
        return False
    else:
        return True


def get_header_lines(
    input_file: str | os.PathLike | io.BytesIO,
    encoding="utf-8",
) -> tuple[list[str], int, int, int]:
    """
    Extract header lines from a file or file-like object.

    This function reads a file or file-like object and extracts lines that
    start with a specific header prefix ("# ") while skipping blank lines
    and lines starting with a comment prefix ("##"). The function stops
    reading at the first non-blank, non-comment line that does not match
    the header prefix.

    Parameters
    ----------
    input_file : str | os.PathLike | io.BytesIO
        The input file path or file-like object to read. If a file path is
        provided, the function automatically handles compressed file formats
        that are supported by `~astropy.utils.data.get_readable_fileobj`.
    encoding : str, optional
        The encoding used to decode the file content. Default is "utf-8".

    Returns
    -------
    lines : list[str]
      List of decoded header lines without the header prefix.
    idx : int
        Index of the last line read.
    n_empty : int
        Number of empty lines read.
    n_comment : int
        Number of comment lines read.
    """
    header_prefix = "# ".encode(encoding)
    comment_prefix = "##".encode(encoding)
    lines = []
    n_empty = 0
    n_comment = 0

    with get_readable_fileobj(input_file, encoding="binary") as f:
        for idx, line in enumerate(f):
            line_strip = line.strip()
            if line_strip.startswith(header_prefix):
                lines.append(line_strip[2:].decode(encoding))
            elif not line_strip:
                n_empty += 1
            elif line_strip.startswith(comment_prefix):
                n_comment += 1
            else:
                # Stop iterating on first failed comment match for a non-blank line
                break

    # Need to rewind the input file if it is a file-like object
    if isinstance(input_file, io.BytesIO):
        input_file.seek(0)

    return lines, idx, n_empty, n_comment


def get_csv_np_type_dtype_shape(
    datatype: str, subtype: str | None, name: str
) -> DerivedColumnProperties:
    """Get the csv_np_type, dtype, and shape of the column from datatype and subtype.

    This function implements most of the complexity of the ECSV data type handling. The
    ECSV standard allows for a wide variety of data types and subtypes, and we also need
    to handle some legacy cases and be liberal in what we accept.

    Parameters
    ----------
    datatype : str
        The data type of the column as specified in the ECSV header.
    subtype : str or None
        The subtype of the column, if applicable. This can include additional
        information like array shape or JSON serialization.
    name : str
        The name of the column, used for error messages.

    Returns
    -------
    CSVNpTypeDtypeShape
        A named tuple containing:
        - `csv_np_type`: Numpy type string for the CSV column data.
        - `dtype`: Numpy dtype in the final column data.
        - `shape`: Shape of the final column data as a tuple of integers.

    Raises
    ------
    ValueError
        If the `datatype` or `subtype` is not recognized or cannot be converted to a
        valid numpy dtype.
    InconsistentTableError
        If the `datatype` is not in the allowed ECSV datatypes and cannot be parsed as a
        numpy dtype.
    """
    from astropy.io.ascii.core import InconsistentTableError
    from astropy.io.ascii.ecsv import ECSV_DATATYPES, InvalidEcsvDatatypeWarning

    csv_np_type = "str" if datatype == "string" else datatype
    dtype = csv_np_type
    shape = ()

    if datatype not in ECSV_DATATYPES:
        msg = (
            f"unexpected datatype {datatype!r} of column {name!r} "
            f"is not in allowed ECSV datatypes {ECSV_DATATYPES}."
        )
        # Try being liberal on input if the `csv_np_type` (derived from ECSV
        # `datatype`) looks like a numpy dtype. In this case, parse the column as
        # string and then cast as `csv_np_type`. This allows for back-compatibility
        # with early versions of io.ascii.ecsv that wrote and read e.g.
        # datatype=datetime64.
        if is_numpy_dtype(csv_np_type):
            dtype = csv_np_type
            csv_np_type = "str"
            warnings.warn(msg, InvalidEcsvDatatypeWarning)
        else:
            # No joy, this is an exception
            raise InconsistentTableError(msg)

    if subtype and csv_np_type != "str":
        # Note: the "column .. failed to convert" bit is odd here but it is to match
        # the io.ascii.ecsv behavior.
        raise ValueError(
            f"column {name!r} failed to convert: "
            f'datatype of column {name!r} must be "string"'
        )

    if subtype:
        # Subtype can be written like "int64[2,null]" and we want to split this
        # out to "int64" and [2, None].
        if "[" in subtype:
            idx = subtype.index("[")
            dtype = subtype[:idx]
            shape = tuple(json.loads(subtype[idx:]))
        else:
            dtype = subtype

        # Map ECSV types to numpy dtypes
        dtype = {"json": "object", "string": "str"}.get(dtype, dtype)

        # Check if the subtype corresponds to a valid numpy dtype. This is required by
        # the astropy implementation, but not by the ECSV standard. The standard states
        # that an unknown subtype can be ignored, so that is what we do here (but with
        # a warning).
        if not is_numpy_dtype(dtype):
            warnings.warn(
                f"unexpected subtype {subtype!r} set for column "
                f"{name!r}, using dtype={csv_np_type!r} instead.",
                category=InvalidEcsvDatatypeWarning,
            )
            dtype = csv_np_type

    return DerivedColumnProperties(csv_np_type, dtype, shape)


def read_header(
    input_file: str | os.PathLike | io.BytesIO,
    encoding: str = "utf-8",
) -> ECSVHeader:
    """
    Read and parse the header of an ECSV (Enhanced Character Separated Values) input.

    This function extracts and validates the ECSV header from the given input file,
    parses the YAML metadata, and constructs the corresponding ECSVHeader object
    containing column definitions and table metadata.

    Parameters
    ----------
    input_file : str, os.PathLike, or io.BytesIO
        The path to the ECSV file or a file-like object containing the ECSV data.
    encoding : str, optional
        The encoding to use when reading the file. Default is 'utf-8'.

    Returns
    -------
    ECSVHeader
        An object containing header information, including the number of header lines,
        number of empty lines, column attributes, table metadata, and delimiter.

    Raises
    ------
    InconsistentTableError
        If the ECSV header is missing, malformed, or the YAML metadata cannot be parsed.
    ValueError
        If the delimiter specified in the header is not supported.

    Notes
    -----
    The function expects the first non-blank comment line to be the ECSV version header,
    and only space and comma are allowed as delimiters in the ECSV format.
    """
    from astropy.io.ascii.core import InconsistentTableError
    from astropy.io.ascii.ecsv import DELIMITERS
    from astropy.table import meta

    # Extract non-blank comment (header) lines with comment character stripped
    header_lines, n_header, n_empty, n_comment = get_header_lines(
        input_file, encoding=encoding
    )

    # Validate that this is a ECSV file
    ecsv_header_re = r"""%ECSV [ ]
                            (?P<major> \d+)
                            \. (?P<minor> \d+)
                            \.? (?P<bugfix> \d+)? $"""

    no_header_msg = (
        'ECSV header line like "# %ECSV <version>" not found as first line.'
        "  This is required for a ECSV file."
    )

    if not header_lines:
        raise InconsistentTableError(no_header_msg)

    match = re.match(ecsv_header_re, header_lines[0].strip(), re.VERBOSE)
    if not match:
        raise InconsistentTableError(no_header_msg)

    try:
        header = meta.get_header_from_yaml(header_lines)
    except meta.YamlParseError as e:
        raise InconsistentTableError("unable to parse yaml in meta header") from e

    table_meta = header.get("meta", None)

    delimiter = header.get("delimiter", " ")
    if delimiter not in DELIMITERS:
        raise ValueError(
            "only space and comma are allowed for delimiter in ECSV format"
        )

    # Create list of columns from `header`.
    cols_attrs = [ColumnECSV(**col) for col in header["datatype"]]

    return ECSVHeader(n_header, n_empty, n_comment, cols_attrs, table_meta, delimiter)


def read_data(
    input_file: str | os.PathLike | io.BytesIO,
    header: ECSVHeader,
    null_values: list[str],
    encoding: str = "utf-8",
    engine_name: str = "io.ascii",
) -> "Table":
    """
    Read the data from an ECSV table using the specified engine.

    This function uses an engine-specific class to handle reading and converting
    the data according to the ECSV specification and the selected backend.

    Parameters
    ----------
    input_file : str, os.PathLike, or io.BytesIO
        The path to the input file or a file-like object containing the ECSV data.
    header : ECSVHeader
        The parsed ECSV header containing column definitions and metadata.
    null_values : list of str
        List of string values to interpret as null/missing values in the data.
    encoding : str, optional
        The encoding to use when reading the file. Default is "utf-8".
    engine_name: str, optional
        The backend engine to use for reading the data. Default is "io.ascii".
        Built-in options are "pyarrow", "pandas", and "io.ascii".

    Returns
    -------
    Table
        An Astropy Table containing the data read from the ECSV file.

    Raises
    ------
    InconsistentTableError
        If the column names from the ECSV header do not match the column names
        in the data.
    """
    from astropy.table import Table

    engine = ECSVEngine.engines[engine_name]()

    # Get the engine-specific kwargs for reading the CSV data.
    kwargs = engine.get_data_kwargs(header, null_values)

    data = Table.read(
        input_file,
        format=engine.format,
        delimiter=header.delimiter,
        encoding=encoding,
        **kwargs,
    )

    # Ensure ECSV header names match the data column names.
    ecsv_header_names = [col.name for col in header.cols]
    if ecsv_header_names != data.colnames:
        from astropy.io.ascii.core import InconsistentTableError

        raise InconsistentTableError(
            f"column names from ECSV header {ecsv_header_names} do not "
            f"match names from header line of CSV data {data.colnames}"
        )

    return data


def get_str_vals(
    data: np.ndarray | np.ma.MaskedArray,
) -> tuple[list[str] | npt.NDArray[np.str_], npt.NDArray[np.bool_] | None]:
    """Get the string values and the mask if available.

    This assumes a 1-d input array of strings, possibly masked. This array comes from
    reading the ECSV data, which is always a 1-d array. This function is only called if
    that array is a numpy string array or a masked array of strings.

    For a masked array it converts the data to the equivalent Python representation
    (list of strings) and returns the mask as a separate array.

    A list of strings is required in this case because the subsequent
    ``process_*_data`` functions substitute (in-place) a new string with the appropriate
    JSON for an empty/masked fill value. In particular, if the original input consists
    solely of empty strings (which is legal), the numpy string array will be not be wide
    enough to hold the fill value.

    For regular numpy arrays it simply returns the original data as a numpy array.

    Parameters
    ----------
    data : np.ndarray | np.ma.MaskedArray
        The input data array to extract string values from.

    Returns
    -------
    str_vals : list[str] | npt.NDArray[np.str_]
        A list of strings or a 1D numpy array of strings representing the data.
    mask : npt.NDArray[np.bool_] | None
        A 1D numpy array of booleans indicating the mask, or None if not applicable.
    """
    # For masked we need a list because for multidim the data under the mask is set
    # to a compatible value.
    if hasattr(data, "mask"):
        # TODO: for not NUMPY_LT_2_0, try changing this to:
        # str_vals = data.astype("T")
        str_vals = data.view(np.ndarray).tolist()
        mask = data.mask
    else:
        str_vals = data
        mask = None
    return str_vals, mask


def convert_column(
    col: ColumnECSV,
    data_in: np.ndarray | np.ma.MaskedArray,
) -> np.ndarray | np.ma.MaskedArray:
    """
    Convert column data from original CSV numpy type to specified output numpy dtype.

    This function handles both regular scalar columns and more complex columns such as:

    - Object dtype columns containing arbitrary Python objects serializable to JSON.
    - Variable-length array columns, where the last axis may vary in length.
    - Fixed-shape multidimensional columns.

    Depending on the column's dtype and shape, the function selects the appropriate
    processing routine to convert the data, including deserialization from JSON where
    necessary. For regular scalar columns, it casts the data to the target dtype if
    needed.

    Parameters
    ----------
    col : ColumnECSV
        The column specification, including dtype, shape, and name.
    data_in : np.ndarray | np.ma.MaskedArray
        The input data array to be converted.

    Returns
    -------
    np.ndarray | np.ma.MaskedArray
        The converted data array with the appropriate dtype and shape.

    Raises
    ------
    ValueError
        If the data cannot be converted due to shape mismatch or invalid JSON content.
    """
    try:
        if col.dtype == "object" or col.shape:
            # Handle three distinct column types where each row element is serialized to
            # JSON. In this case ``data_in`` is an ndarray or MaskedArray of
            # fixed-length string which are the JSON-encoded representation of the data.
            # See docstring in `get_str_vals` for explanation of the next step, which
            # has some subtlety.
            str_vals, mask = get_str_vals(data_in)

            if col.dtype == "object":
                # Any Python objects serializable to JSON
                process_func = process_object_data
            elif col.shape[-1] is None:
                # Variable length arrays with shape (n, m, ..., *) for fixed
                # n, m, .. and variable in last axis.
                process_func = process_variable_length_array_data
            else:
                # Multidim columns with consistent shape (n, m, ...).
                process_func = process_fixed_shape_multidim_data

            data_out, col_shape = process_func(col, str_vals, mask)

        # Regular scalar value column
        else:
            data_out = data_in
            # If we need to cast the data to a different dtype, do it now.
            if data_out.dtype != col.dtype:
                data_out = data_out.astype(col.dtype)
            col_shape = col.shape

        if data_out.shape[1:] != tuple(col_shape):
            raise ValueError("shape mismatch between value and column specifier")

    except json.JSONDecodeError:
        raise ValueError(
            f"column {col.name!r} failed to convert: column value is not valid JSON"
        )
    except Exception as exc:
        raise ValueError(f"column {col.name!r} failed to convert: {exc}") from exc

    return data_out


def process_object_data(
    col: ColumnECSV,
    str_vals: list[str] | npt.NDArray[np.str_],
    mask: npt.NDArray[np.bool_] | None,
) -> tuple[np.ndarray | np.ma.MaskedArray, tuple[int, ...]]:
    """
    Handle object columns where each row element is a JSON-encoded object.

    The ECSV format only allows a 1-d column of object type.

    Example::

        # %ECSV 1.0
        # ---
        # datatype:
        # - {name: objects, datatype: string, subtype: json}
        # schema: astropy-2.0
        objects
        "{""a"":1}"
        "{""b"":[2.5,null]}"
        true

    Parameters
    ----------
    col : ColumnECSV
        The column specification, including dtype, shape, and name.
    str_vals : list[str] or 1-D ndarray[str]
        JSON-encoded string representations of the data.
    mask : 1-D ndarray[bool] or None
        Boolean mask array 1-D indicating invalid or missing values. If None, no masking
        is applied.

    Returns
    -------
    data_out : numpy.ndarray or numpy.ma.MaskedArray
        An array of objects reconstructed from `str_vals`, with the same shape as `col`.
        If `mask` is provided, a masked array is returned with the mask applied.
    col_shape : tuple[int, ...]
        Expected shape of data_out, used in final sanity check of reading.
    """
    if mask is not None:
        for idx in np.nonzero(mask)[0]:
            str_vals[idx] = "0"  # could be "null" but io.ascii uses "0"
    col_vals = [json.loads(val) for val in str_vals]
    np_empty = np.empty if mask is None else np.ma.empty
    data_out = np_empty((len(col_vals),) + tuple(col.shape), dtype=object)
    data_out[...] = col_vals
    if mask is not None:
        data_out.mask = mask
    return data_out, col.shape


def process_fixed_shape_multidim_data(
    col: ColumnECSV,
    str_vals: list[str] | npt.NDArray[np.str_],
    mask: npt.NDArray[np.bool_] | None,
) -> tuple[np.ndarray | np.ma.MaskedArray, tuple[int, ...]]:
    """
    Handle fixed-shape multidimensional columns as JSON-encoded strings.

    Example::

        # %ECSV 1.0
        # ---
        # datatype:
        # - {name: array3x2, datatype: string, subtype: 'float64[3,2]'}
        # schema: astropy-2.0
        array3x2
        [[0.0,1.0],[2.0,3.0],[4.0,5.0]]
        [[6.0,7.0],[8.0,null],[10.0,11.0]]

    Parameters
    ----------
    col : ColumnECSV
        The column specification, including dtype, shape, and name.
    str_vals : array-like of str
        Array of string representations of the data, typically JSON-encoded.
    mask : numpy.ndarray or None
        Boolean mask array indicating invalid or missing values. If None, no masking is
        applied.

    Returns
    -------
    data_out : numpy.ndarray or numpy.ma.MaskedArray
        An array of objects reconstructed from `str_vals`, with the same shape as `col`.
        If `mask` is provided, a masked array is returned with the mask applied.
    col_shape : tuple[int, ...]
        Expected shape of data_out, used in final sanity check of reading.
    """
    # Change empty (blank) values in original ECSV to something
    # like "[[null, null],[null,null]]" so subsequent JSON
    # decoding works.
    if mask is not None:
        all_none_arr = np.full(shape=col.shape, fill_value=None, dtype=object)
        fill_value = json.dumps(all_none_arr.tolist())
        for idx in np.nonzero(mask)[0]:
            str_vals[idx] = fill_value

    col_vals = [json.loads(val) for val in str_vals]

    # Make a numpy object array of col_vals to look for None (masked values)
    arr_vals = np.array(col_vals, dtype=object)
    arr_vals_mask = arr_vals == None
    if np.any(arr_vals_mask):
        # Replace all the None with an appropriate fill value
        arr_vals[arr_vals_mask] = {"U": "", "S": b""}.get(col.dtype.kind, 0)
        # Finally make a MaskedArray with the filled data + mask
        data_out = np.ma.array(arr_vals.astype(col.dtype), mask=arr_vals_mask)
    else:
        data_out = arr_vals.astype(col.dtype)

    return data_out, col.shape


def process_variable_length_array_data(
    col: ColumnECSV,
    str_vals: list[str] | npt.NDArray[np.str_],
    mask: npt.NDArray[np.bool_] | None,
) -> tuple[np.ndarray | np.ma.MaskedArray, tuple[int, ...]]:
    """
    Handle variable length arrays with shape (n, m, ..., *) as JSON-encoded strings.

    Shape is fixed for n, m, .. and variable in last axis. The output is a 1-d object
    array with each row element being an ``np.ndarray`` or ``np.ma.masked_array`` of the
    appropriate shape.

    Example::

        # %ECSV 1.0
        # ---
        # datatype:
        # - {name: array_var, datatype: string, subtype: 'int64[null]'}
        # schema: astropy-2.0
        array_var
        [1,2]
        [3,4,5,null,7]
        [8,9,10]

    Parameters
    ----------
    col : ColumnECSV
        The column specification, including dtype, shape, and name.
    str_vals : list[str] or 1-D ndarray[str]
        JSON-encoded string representations of the data.
    mask : 1-D ndarray[bool] or None
        Boolean mask array 1-D indicating invalid or missing values. If None, no masking
        is applied.

    Returns
    -------
    data_out : numpy.ndarray or numpy.ma.MaskedArray
        An array of objects reconstructed from `str_vals`, with the same shape as `col`.
        If `mask` is provided, a masked array is returned with the mask applied.
    col_shape : tuple[int, ...]
        Expected shape of data_out, used in final sanity check of reading.
    """
    # Empty (blank) values in original ECSV are masked. Instead set the values
    # to "[]" indicating an empty list. This operation also unmasks the values.
    if mask is not None:
        fill_value = "[]"
        for idx in np.nonzero(mask)[0]:
            str_vals[idx] = fill_value

    # Remake as a 1-d object column of numpy ndarrays or
    # MaskedArray using the datatype specified in the ECSV file.
    col_vals = []
    for str_val in str_vals:
        obj_val = json.loads(str_val)  # list or nested lists
        try:
            arr_val = np.array(obj_val, dtype=col.dtype)
        except TypeError:
            # obj_val has entries that are inconsistent with
            # dtype. For a valid ECSV file the only possibility
            # is None values (indicating missing values).
            vals = np.array(obj_val, dtype=object)
            # Replace all the None with an appropriate fill value
            mask_vals = vals == None
            vals[mask_vals] = {"U": "", "S": b""}.get(col.dtype.kind, 0)
            arr_val = np.ma.array(vals.astype(col.dtype), mask=mask_vals)

        col_vals.append(arr_val)

    np_empty = np.empty if mask is None else np.ma.empty
    data_out = np_empty(len(col_vals), dtype=object)
    data_out[:] = col_vals
    if mask is not None:
        data_out.mask = mask
    return data_out, ()


def get_null_values_per_column(
    cols: list[ColumnECSV],
    table_meta: dict | None,
    null_values: list[str],
) -> list[tuple[str, str, str]]:
    """Get null and fill values for individual columns.

    For ECSV to handle the corner case of data that has been serialized using the
    serialize_method='data_mask' option, which writes the full data and mask directly,
    AND where that table includes a string column with zero-length string entries ("")
    which are valid data. Normally the super() method will set col.fill_value=('', '0')
    to replace blanks with a '0'.  But for that corner case subset, instead do not do
    any filling.

    Parameters
    ----------
    cols : list[ColumnECSV]
        List of ColumnECSV objects representing the columns in the ECSV file.
    table_meta : dict or None
        Metadata dictionary from the ECSV header, which may include serialized columns
        and other metadata.
    null_values : list[str]
        List of string values to interpret as null/missing values in every column.
        The upstream default from ``read_ecsv`` is [""] but no default is defined here.

    Returns
    -------
    fill_values : list[tuple[str, str, str]]
        A list of tuples with (null_value, fill_value, column_name) for each column in
        `cols` that is not a MaskedColumn. If no fill values are needed, returns an
        empty list.
    """
    if table_meta is None:
        table_meta = {}

    # Get the serialized columns spec or an empty dict if not present.
    serialized_columns: dict[str, SerializedColumn] = table_meta.get(
        "__serialized_columns__", {}
    )

    # A serialized MaskedColumn column (via `serialize_method="data_mask"`) does not
    # have a fill value, so assemble a set of columns names to skip include the data and
    # the mask columns. For example:
    # - __serialized_columns__:
    #     a:
    #       __class__: astropy.table.column.MaskedColumn
    #       data: !astropy.table.SerializedColumn {name: a}
    #       mask: !astropy.table.SerializedColumn {name: a.mask}
    masked_col_names = set()
    for name, sc in serialized_columns.items():
        if sc["__class__"] == "astropy.table.column.MaskedColumn":
            masked_col_names.add(name)
            masked_col_names.add(name + ".mask")

    fill_values = []
    for col in cols:
        if col.name in masked_col_names:
            continue

        fill_value = "" if col.csv_np_type == "str" else "0"
        for null_value in null_values:
            fill_values.append((null_value, fill_value, col.name))

    return fill_values


def read_ecsv(
    input_file: str | os.PathLike | io.BytesIO | io.StringIO | Iterable[str],
    *,
    encoding: str = "utf-8",
    engine: str = "io.ascii",
    null_values: list[str] | None = None,
) -> "Table":
    """
    Read an ECSV (Enhanced Character Separated Values) file and return an Astropy Table.

    Parameters
    ----------
    input_file : str, os.PathLike, io.BytesIO, io.StringIO, Iterable[str]
        The ECSV input to read. This can be a file path, a file-like object, a string
        containing the file contents, or an iterable of strings representing lines of
        the file. Note that providing ``io.StringIO`` or an iterable of strings will be
        less memory efficient, as it will be converted to a bytes stream.
    encoding : str, optional
        The encoding to use when reading the file. Default is "utf-8".
    engine : str, optional
        The engine to use for reading the CSV data. Default is "io.ascii", which uses
        astropy to read the CSV data. Other built-in options are "pyarrow" and "pandas".
        The "pyarrow" engine is optimized for performance and can handle large datasets
        efficiently. The "pandas" engine uses the pandas CSV reader, which is also
        faster than the default "io.ascii" engine.
    null_values : list of str or None, optional
        List of string values to interpret as null/missing values. Default is [""]. The
        ECSV standard requires the null values are represented as empty strings in the
        CSV data, but this allows reading non-compliant ECSV files. A notable example
        are the Gaia source download files which are ECSV but use "null".

    Returns
    -------
    table : astropy.table.Table
        The table read from the ECSV file.

    Raises
    ------
    astropy.io.ascii.core.InconsistentTableError
        If the column names in the ECSV header do not match the column names in the CSV
        data.

    Notes
    -----
    - The function handles various input types, including file paths, file-like objects,
      and in-memory strings or lists of strings.
    - Metadata and column attributes (such as unit, description, format, and meta) are
      transferred from the ECSV header to the resulting Table object.
    - Handles JSON-encoded data and ensures appropriate numpy dtypes for columns.
    """
    from astropy.table import Table, serialize

    if null_values is None:
        null_values = [""]

    # Allow input types that are historically supported by io.ascii. These will not be
    # memory or speed efficient but will still work.
    if isinstance(input_file, io.StringIO):
        input_file = io.BytesIO(input_file.getvalue().encode(encoding))
    elif isinstance(input_file, str) and "\n" in input_file:
        input_file = io.BytesIO(input_file.encode(encoding))
    elif isinstance(input_file, (list, tuple)):
        # TODO: better way to check for an iterable of str?
        input_file = io.BytesIO("\n".join(input_file).encode(encoding))

    # Read the ECSV header from the input.
    header = read_header(input_file, encoding=encoding)

    # Read the CSV data from the input starting at the line after the header. This
    # includes handling that is particular to the engine.
    data_raw = read_data(
        input_file,
        header,
        null_values=null_values,
        encoding=encoding,
        engine_name=engine,
    )

    # Convert the column data to the appropriate numpy dtype. This is mostly concerned
    # with JSON-encoded data but also handles cases like pyarrow not supporting float16.
    data = {col.name: convert_column(col, data_raw[col.name]) for col in header.cols}

    # Create the Table object
    table = Table(data)

    # Transfer metadata from the ECSV header to the Table columns.
    for header_col in header.cols:
        col = table[header_col.name]
        for attr in ["unit", "description", "format", "meta"]:
            if (val := getattr(header_col, attr)) is not None:
                setattr(col.info, attr, val)

    # Add metadata to the table
    if header.table_meta:
        table.meta.update(header.table_meta)

    # Construct any mixin columns from the raw columns.
    table = serialize._construct_mixins_from_columns(table)

    return table  # noqa: RET504


def write_ecsv(tbl, output, engine="io.ascii", **kwargs):
    """Thin wrapper around the ``io.ascii`` ECSV writer to write ECSV files.

    Parameters
    ----------
    tbl : astropy.table.Table
        The table to write to ECSV format.
    output : str or os.PathLike or file-like object
        The output file path or file-like object to write the ECSV data to.
    engine : str, optional
        The engine to use for writing the CSV data. Default is "io.ascii", which uses
        astropy to write the CSV data. Currently this is the only option.
    **kwargs : dict, optional
        Additional keyword arguments passed to the ECSV writer. These can include
        options like ``delimiter``, ``encoding``, and others supported by the
        `astropy.io.ascii.Ecsv` writer.
    """
    if engine != "io.ascii":
        raise ValueError(
            f"{engine=} is not a supported engine for writing, use 'io.ascii'"
        )
    tbl.write(output, format="ascii.ecsv", **kwargs)


def register_ecsv_table():
    """
    Register ECSV reader and writer with Unified I/O as a Table reader.
    """
    from astropy.io import registry as io_registry
    from astropy.table import Table

    io_registry.register_reader("ecsv", Table, read_ecsv)
    io_registry.register_writer("ecsv", Table, write_ecsv)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

from unittest.mock import MagicMock, patch

import matplotlib.transforms as transforms
import pytest
from matplotlib.backends.backend_agg import FigureCanvasAgg
from matplotlib.figure import Figure

from astropy import units as u
from astropy.io import fits
from astropy.utils.data import get_pkg_data_filename
from astropy.utils.exceptions import AstropyDeprecationWarning
from astropy.visualization.wcsaxes.coordinate_helpers import CoordinateHelper
from astropy.visualization.wcsaxes.core import WCSAxes
from astropy.wcs import WCS

MSX_HEADER = fits.Header.fromtextfile(get_pkg_data_filename("data/msx_header"))


def test_getaxislabel(ignore_matplotlibrc):
    fig = Figure()
    ax = WCSAxes(fig, [0.1, 0.1, 0.8, 0.8], aspect="equal")

    ax.coords[0].set_axislabel("X")
    ax.coords[1].set_axislabel("Y")
    assert ax.coords[0].get_axislabel() == "X"
    assert ax.coords[1].get_axislabel() == "Y"


@pytest.fixture
def ax():
    fig = Figure()
    _canvas = FigureCanvasAgg(fig)
    ax = WCSAxes(fig, [0.1, 0.1, 0.8, 0.8], aspect="equal")
    fig.add_axes(ax)

    return ax


def assert_label_draw(ax, x_label, y_label):
    ax.coords[0].set_axislabel("Label 1")
    ax.coords[1].set_axislabel("Label 2")

    with patch.object(ax.coords[0]._axislabels, "set_position") as pos1:
        with patch.object(ax.coords[1]._axislabels, "set_position") as pos2:
            ax.figure.canvas.draw()

    assert pos1.call_count == x_label
    assert pos2.call_count == y_label


def test_label_visibility_rules_default(ignore_matplotlibrc, ax):
    assert_label_draw(ax, True, True)


def test_label_visibility_rules_label(ignore_matplotlibrc, ax):
    ax.coords[0].set_ticklabel_visible(False)
    ax.coords[1].set_ticks(values=[-9999] * u.one)

    assert_label_draw(ax, False, False)


def test_label_visibility_rules_ticks(ignore_matplotlibrc, ax):
    ax.coords[0].set_axislabel_visibility_rule("ticks")
    ax.coords[1].set_axislabel_visibility_rule("ticks")

    ax.coords[0].set_ticklabel_visible(False)
    ax.coords[1].set_ticks(values=[-9999] * u.one)

    assert_label_draw(ax, True, False)


def test_label_visibility_rules_always(ignore_matplotlibrc, ax):
    ax.coords[0].set_axislabel_visibility_rule("always")
    ax.coords[1].set_axislabel_visibility_rule("always")

    ax.coords[0].set_ticklabel_visible(False)
    ax.coords[1].set_ticks(values=[-9999] * u.one)

    assert_label_draw(ax, True, True)


def test_format_unit():
    fig = Figure()
    canvas = FigureCanvasAgg(fig)
    ax = WCSAxes(fig, [0.1, 0.1, 0.8, 0.8], wcs=WCS(MSX_HEADER))
    fig.add_axes(ax)

    # Force a draw which is required for format_coord to work
    canvas.draw()

    ori_fu = ax.coords[1].get_format_unit()
    assert ori_fu == "deg"

    ax.coords[1].set_format_unit("arcsec")
    fu = ax.coords[1].get_format_unit()
    assert fu == "arcsec"


def test_set_separator():
    fig = Figure()
    canvas = FigureCanvasAgg(fig)
    ax = WCSAxes(fig, [0.1, 0.1, 0.8, 0.8], wcs=WCS(MSX_HEADER))
    fig.add_axes(ax)

    # Force a draw which is required for format_coord to work
    canvas.draw()

    ax.coords[1].set_format_unit("deg")
    assert ax.coords[1].format_coord(4) == "4\xb000'00\""
    ax.coords[1].set_separator((":", ":", ""))
    assert ax.coords[1].format_coord(4) == "4:00:00"
    ax.coords[1].set_separator("abc")
    assert ax.coords[1].format_coord(4) == "4a00b00c"
    ax.coords[1].set_separator(None)
    assert ax.coords[1].format_coord(4) == "4\xb000'00\""


@pytest.mark.parametrize(
    "draw_grid, expected_visibility", [(True, True), (False, False), (None, True)]
)
def test_grid_variations(ignore_matplotlibrc, draw_grid, expected_visibility):
    fig = Figure()
    ax = WCSAxes(fig, [0.1, 0.1, 0.8, 0.8], aspect="equal")
    fig.add_axes(ax)
    transform = transforms.Affine2D().scale(2.0)
    coord_helper = CoordinateHelper(
        parent_axes=ax,
        transform=transform,
        frame=MagicMock(),
    )
    coord_helper.grid(draw_grid=draw_grid)
    assert coord_helper._grid_lines_kwargs["visible"] == expected_visibility


def test_get_position():
    fig = Figure()
    _canvas = FigureCanvasAgg(fig)
    ax = WCSAxes(fig, [0.1, 0.1, 0.8, 0.8], aspect="equal")
    fig.add_axes(ax)

    assert ax.coords[0].get_ticks_position() == ["b", "r", "t", "l"]
    assert ax.coords[1].get_ticks_position() == ["b", "r", "t", "l"]
    assert ax.coords[0].get_ticklabel_position() == ["#"]
    assert ax.coords[1].get_ticklabel_position() == ["#"]
    assert ax.coords[0].get_axislabel_position() == ["#"]
    assert ax.coords[1].get_axislabel_position() == ["#"]

    fig.canvas.draw()

    assert ax.coords[0].get_ticks_position() == ["b", "r", "t", "l"]
    assert ax.coords[1].get_ticks_position() == ["b", "r", "t", "l"]
    assert ax.coords[0].get_ticklabel_position() == ["b", "#"]
    assert ax.coords[1].get_ticklabel_position() == ["l", "#"]
    assert ax.coords[0].get_axislabel_position() == ["b", "#"]
    assert ax.coords[1].get_axislabel_position() == ["l", "#"]

    ax.coords[0].set_ticks_position("br")
    ax.coords[1].set_ticks_position("tl")
    ax.coords[0].set_ticklabel_position("bt")
    ax.coords[1].set_ticklabel_position("rl")
    ax.coords[0].set_axislabel_position("t")
    ax.coords[1].set_axislabel_position("r")

    assert ax.coords[0].get_ticks_position() == ["b", "r"]
    assert ax.coords[1].get_ticks_position() == ["t", "l"]
    assert ax.coords[0].get_ticklabel_position() == ["b", "t"]
    assert ax.coords[1].get_ticklabel_position() == ["r", "l"]
    assert ax.coords[0].get_axislabel_position() == ["t"]
    assert ax.coords[1].get_axislabel_position() == ["r"]


def test_deprecated_getters():
    fig = Figure()
    _canvas = FigureCanvasAgg(fig)
    ax = WCSAxes(fig, [0.1, 0.1, 0.8, 0.8], aspect="equal")
    fig.add_axes(ax)

    helper = CoordinateHelper(
        parent_axes=ax,
        frame=MagicMock(),
    )

    with pytest.warns(AstropyDeprecationWarning):
        ticks = helper.ticks
    assert not ticks.get_display_minor_ticks()
    with pytest.warns(AstropyDeprecationWarning):
        ticklabels = helper.ticklabels
    assert ticklabels.text == {}
    with pytest.warns(AstropyDeprecationWarning):
        axislabels = helper.axislabels
    assert axislabels.get_visibility_rule() == "labels"


def test_set_major_formatter():
    fig = Figure()
    canvas = FigureCanvasAgg(fig)
    ax = WCSAxes(fig, [0.1, 0.1, 0.8, 0.8], wcs=WCS(MSX_HEADER))
    fig.add_axes(ax)

    # Force a draw which is required for format_coord to work
    canvas.draw()

    ax.coords[1].set_major_formatter("d.ddd")
    assert ax.coords[1].format_coord(4) == "4.000\xb0"

    ax.coords[1].set_major_formatter("d.dd", show_decimal_unit=False)
    assert ax.coords[1].format_coord(4) == "4.00"

    # Show unit has no effect on sexagesimal coordinates

    ax.coords[1].set_major_formatter("dd:mm:ss.s", show_decimal_unit=False)
    assert ax.coords[1].format_coord(4) == "4\xb000'00.0\""


def test_get_ticks_visible():
    fig = Figure()
    canvas = FigureCanvasAgg(fig)
    ax = WCSAxes(fig, [0.1, 0.1, 0.8, 0.8], wcs=WCS(MSX_HEADER))
    fig.add_axes(ax)

    assert ax.coords[0].get_ticks_visible() is True

    ax.coords[0].set_ticks_visible(False)
    assert ax.coords[0].get_ticks_visible() is False

    ax.coords[0].set_ticks_visible(True)
    assert ax.coords[0].get_ticks_visible() is True


def test_get_ticklabel_visible():
    fig = Figure()
    canvas = FigureCanvasAgg(fig)
    ax = WCSAxes(fig, [0.1, 0.1, 0.8, 0.8], wcs=WCS(MSX_HEADER))
    fig.add_axes(ax)

    assert ax.coords[0].get_ticklabel_visible() is True

    ax.coords[0].set_ticklabel_visible(False)
    assert ax.coords[0].get_ticklabel_visible() is False

    ax.coords[0].set_ticklabel_visible(True)
    assert ax.coords[0].get_ticklabel_visible() is True


def test_get_axislabel_visible():
    fig = Figure()
    canvas = FigureCanvasAgg(fig)
    ax = WCSAxes(fig, [0.1, 0.1, 0.8, 0.8], wcs=WCS(MSX_HEADER))
    fig.add_axes(ax)

    assert ax.coords[0].get_axislabel_visible() is True

    ax.coords[0]._axislabels.set_visible(False)
    assert ax.coords[0].get_axislabel_visible() is False

    ax.coords[0]._axislabels.set_visible(True)
    assert ax.coords[0].get_axislabel_visible() is True


def test_set_visible():
    fig = Figure()
    canvas = FigureCanvasAgg(fig)
    ax = WCSAxes(fig, [0.1, 0.1, 0.8, 0.8], wcs=WCS(MSX_HEADER))
    fig.add_axes(ax)

    assert ax.coords[0].get_ticks_visible() is True
    assert ax.coords[0].get_ticklabel_visible() is True
    assert ax.coords[0].get_axislabel_visible() is True

    ax.coords[0].set_visible(False)
    assert ax.coords[0].get_ticks_visible() is False
    assert ax.coords[0].get_ticklabel_visible() is False
    assert ax.coords[0].get_axislabel_visible() is False

    ax.coords[0].set_visible(True)
    assert ax.coords[0].get_ticks_visible() is True
    assert ax.coords[0].get_ticklabel_visible() is True
    assert ax.coords[0].get_axislabel_visible() is True


@pytest.mark.parametrize("invalid_input", [123, 123.45, None, [], {}])
def test_set_visible_invalid_type(invalid_input):
    fig = Figure()
    canvas = FigureCanvasAgg(fig)
    ax = WCSAxes(fig, [0.1, 0.1, 0.8, 0.8], wcs=WCS(MSX_HEADER))
    fig.add_axes(ax)

    with pytest.raises(TypeError):
        ax.coords[0].set_visible(invalid_input)


def test_set_position():
    fig = Figure()
    canvas = FigureCanvasAgg(fig)
    ax = WCSAxes(fig, [0.1, 0.1, 0.8, 0.8], wcs=WCS(MSX_HEADER))
    fig.add_axes(ax)

    ax.coords[0].set_position("t")
    assert ax.coords[0].get_ticks_position() == ["t"]
    assert ax.coords[0].get_ticklabel_position() == ["t"]
    assert ax.coords[0].get_axislabel_position() == ["t"]

    assert ax.coords[0].get_ticks_visible() is True
    assert ax.coords[0].get_ticklabel_visible() is True
    assert ax.coords[0].get_axislabel_visible() is True

    ax.coords[0].set_position("bl")
    assert ax.coords[0].get_ticks_position() == ["b", "l"]
    assert ax.coords[0].get_ticklabel_position() == ["b", "l"]
    assert ax.coords[0].get_axislabel_position() == ["b", "l"]

    assert ax.coords[0].get_ticks_visible() is True
    assert ax.coords[0].get_ticklabel_visible() is True
    assert ax.coords[0].get_axislabel_visible() is True


@pytest.mark.parametrize("invalid_input", [123, 123.45, None, [], {}])
def test_set_position_invalid_type(invalid_input):
    fig = Figure()
    canvas = FigureCanvasAgg(fig)
    ax = WCSAxes(fig, [0.1, 0.1, 0.8, 0.8], wcs=WCS(MSX_HEADER))
    fig.add_axes(ax)

    with pytest.raises(TypeError):
        ax.coords[0].set_position(invalid_input)


def test_set_position_invalid():
    fig = Figure()
    _canvas = FigureCanvasAgg(fig)
    ax = WCSAxes(fig, [0.1, 0.1, 0.8, 0.8], aspect="equal")
    fig.add_axes(ax)

    with pytest.warns(
        AstropyDeprecationWarning,
        match=r"Ignoring unrecognized position\(s\): \['x'\], should be one of b/r/t/l",
    ):
        ax.coords[0].set_ticks_position("xl")
    with pytest.warns(
        AstropyDeprecationWarning,
        match=r"Ignoring unrecognized position\(s\): \['o'\], should be one of b/r/t/l",
    ):
        ax.coords[1].set_ticklabel_position("to")
    with pytest.warns(
        AstropyDeprecationWarning,
        match=r"Ignoring unrecognized position\(s\): \['q', 'p'\], should be one of b/r/t/l",
    ):
        ax.coords[1].set_axislabel_position("qbp")

    assert ax.coords[0].get_ticks_position() == ["l"]
    assert ax.coords[1].get_ticklabel_position() == ["t"]
    assert ax.coords[1].get_axislabel_position() == ["b"]


def test_set_position_invalid_gridline():
    fig = Figure()
    _canvas = FigureCanvasAgg(fig)
    ax = WCSAxes(fig, [0.1, 0.1, 0.8, 0.8], aspect="equal")
    fig.add_axes(ax)

    ax.coords[0].add_tickable_gridline("my-grid-line", -30 * u.one)

    with pytest.warns(
        AstropyDeprecationWarning,
        match=r"Ignoring unrecognized position\(s\): \['my-parrot-line'\]",
    ):
        ax.coords[1].set_ticks_position(["my-grid-line", "my-parrot-line"])

    with pytest.warns(
        AstropyDeprecationWarning,
        match=r"It looks like 'my-grid-line' matches the name of a single axis. If "
        r"you are trying to specify a multi-character axis name, use a list "
        r"or a tuple, e.g. \('my-grid-line',\).",
    ):
        ax.coords[1].set_ticks_position("my-grid-line")


def test_set_ticks_values():
    fig = Figure()
    _canvas = FigureCanvasAgg(fig)
    ax = WCSAxes(fig, [0.1, 0.1, 0.8, 0.8], wcs=WCS(MSX_HEADER))
    fig.add_axes(ax)

    xticks = [1795, 1780, 1783] * u.arcsec
    ax.coords[0].set_format_unit(u.arcsec)
    ax.coords[0].set_ticks(xticks)

    # Force a draw to calculate the tick positions
    _canvas.draw()
    # This attribute only exists after a draw
    lbl_world = ax.coords[0]._lbl_world
    lbl_world1 = lbl_world[: len(lbl_world) // 2]

    lbl_locations = u.Quantity(lbl_world1, unit=u.deg)
    assert u.allclose(lbl_locations, ax.coords[0]._formatter_locator.values)
    assert u.Quantity(lbl_world).unit is xticks.unit



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""Custom angle errors and exceptions used in :mod:`astropy.coordinates.angles`."""

__all__ = [
    "BoundsError",
    "IllegalHourError",
    "IllegalHourWarning",
    "IllegalMinuteError",
    "IllegalMinuteWarning",
    "IllegalSecondError",
    "IllegalSecondWarning",
    "RangeError",
]

from astropy.utils.exceptions import AstropyWarning


class RangeError(ValueError):
    """
    Raised when some part of an angle is out of its valid range.
    """


class BoundsError(RangeError):
    """
    Raised when an angle is outside of its user-specified bounds.
    """


class IllegalHourError(RangeError):
    """
    Raised when an hour value is not in the range [0,24).

    Parameters
    ----------
    hour : int, float

    Examples
    --------
    .. code-block:: python

        if not 0 <= hr < 24:
           raise IllegalHourError(hour)
    """

    def __init__(self, hour):
        self.hour = hour

    def __str__(self):
        return (
            f"An invalid value for 'hours' was found ('{self.hour}'); must be in the"
            " range [0,24)."
        )


class IllegalHourWarning(AstropyWarning):
    """
    Raised when an hour value is 24.

    Parameters
    ----------
    hour : int, float
    """

    def __init__(self, hour, alternativeactionstr=None):
        self.hour = hour
        self.alternativeactionstr = alternativeactionstr

    def __str__(self):
        message = (
            f"'hour' was found  to be '{self.hour}', which is not in range (-24, 24)."
        )
        if self.alternativeactionstr is not None:
            message += " " + self.alternativeactionstr
        return message


class IllegalMinuteError(RangeError):
    """
    Raised when an minute value is not in the range [0,60].

    Parameters
    ----------
    minute : int, float

    Examples
    --------
    .. code-block:: python

        if not 0 <= min < 60:
            raise IllegalMinuteError(minute)

    """

    def __init__(self, minute):
        self.minute = minute

    def __str__(self):
        return (
            f"An invalid value for 'minute' was found ('{self.minute}'); should be in"
            " the range [0,60)."
        )


class IllegalMinuteWarning(AstropyWarning):
    """
    Raised when a minute value is 60.

    Parameters
    ----------
    minute : int, float
    """

    def __init__(self, minute, alternativeactionstr=None):
        self.minute = minute
        self.alternativeactionstr = alternativeactionstr

    def __str__(self):
        message = (
            f"'minute' was found  to be '{self.minute}', which is not in range [0,60)."
        )
        if self.alternativeactionstr is not None:
            message += " " + self.alternativeactionstr
        return message


class IllegalSecondError(RangeError):
    """
    Raised when an second value (time) is not in the range [0,60].

    Parameters
    ----------
    second : int, float

    Examples
    --------
    .. code-block:: python

        if not 0 <= sec < 60:
            raise IllegalSecondError(second)
    """

    def __init__(self, second):
        self.second = second

    def __str__(self):
        return (
            f"An invalid value for 'second' was found ('{self.second}'); should be in"
            " the range [0,60)."
        )


class IllegalSecondWarning(AstropyWarning):
    """
    Raised when a second value is 60.

    Parameters
    ----------
    second : int, float
    """

    def __init__(self, second, alternativeactionstr=None):
        self.second = second
        self.alternativeactionstr = alternativeactionstr

    def __str__(self):
        message = (
            f"'second' was found  to be '{self.second}', which is not in range [0,60)."
        )
        if self.alternativeactionstr is not None:
            message += " " + self.alternativeactionstr
        return message
